import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pandas as pd
from astropy.time import Time

from astrosn.supernova import SN
from astrosn.photometry import Photometry, PhotFactory
from astrosn.utils import StrEnum
from .utils import add_phot, update_sn, verify_columns
from .collators import Collator
from .converters import BaseConverter, Converter
from .readers import PathType, read_astropy_table, read_pandas_csv


@dataclass(frozen=True)
class ZTFFlowsConverter(BaseConverter, Converter):
    glob_str: str = '*ztf.ecsv'
    jd_col: str = 'time'
    mag_col: str = 'mag'
    mag_err_col: str = 'mag_err'
    band_col: str = 'photfilter'
    sub_col: Optional[str] = None
    site_col: Optional[str] = None

    def site_convert(self) -> None:
        self.df['site'] = 0  # refactor to "ZTF" and for phot to use Site class.

    def sub_convert(self) -> None:
        self.df['sub'] = True

    def band_convert(self) -> None:
        self.df['band'] = self.df[self.band_col].str[:1]


def read_ztf_flows(path: PathType) -> pd.DataFrame:
    """
    Read a ZTF flows file and return a DataFrame.
    """
    at = read_astropy_table(path, fmt='ascii.ecsv')
    at[ZTFFlowsConverter.jd_col] = Time(at[ZTFFlowsConverter.jd_col], format='mjd').jd
    return at.to_pandas()


collate_ztf_flows = Collator(ZTFFlowsConverter, read_ztf_flows)


@dataclass(frozen=True)
class ZTFPhotConverter(BaseConverter, Converter):
    glob_str: str = 'SN*_detections.csv'
    jd_col: str = 'jd'
    mag_col: str = 'mag'
    mag_err_col: str = 'mag_err'
    band_col: str = 'filter'
    sub_col: Optional[str] = None
    site_col: Optional[str] = None

    def site_convert(self) -> None:
        self.df['site'] = 0  # refactor to "ZTF" and for phot to use Site class.

    def sub_convert(self) -> None:
        self.df['sub'] = True

    def band_convert(self) -> None:
        self.df['band'] = self.df[self.band_col].str[-1]


@dataclass(frozen=True)
class ZTFPhotLimConverter(ZTFPhotConverter):
    glob_str: str = 'SN*_limits.csv'
    mag_col: str = 'lim'


def read_ztfphot2(path: PathType) -> pd.DataFrame:
    return read_pandas_csv(path, index_col=0)


collate_ztfphot = Collator(ZTFPhotConverter, read_ztfphot2)
collate_ztfphot_limits = Collator(ZTFPhotLimConverter, read_ztfphot2)


def add_ztf_photometry(sn: SN, path: str | Path) -> SN:
    """
    Add ZTF photometry to a Supernova.
    """
    limits = collate_ztfphot_limits(path)
    detections = collate_ztfphot(path)

    new_phot = add_phot(sn.phot, detections)
    sn = update_sn(sn, new_phot)
    new_limits = add_phot(sn.limits, limits)
    sn.limits = new_limits
    return sn

# Below is Deprecated and will be removed in a future release


class ZTFPhotSuffix(StrEnum):
    """
    Photometry from ZTF
    """
    detections = '*_detections.csv'
    limits = '*_limits.csv'


def read_ztfphot_old(path: PathType, ztf_site_id: int = 0) -> pd.DataFrame:
    path = Path(path).resolve(strict=True)  # strict=True raises FileNotFoundError if doesn't exist
    name, filt, phot_type = path.stem.split('_')[:3]
    df = pd.read_csv(path, index_col=0)

    # Modify the columns to match the standard
    if phot_type == 'limits':
        df['mag'] = df['lim']
    df['site'] = ztf_site_id
    df['band'] = filt
    df['sub'] = True
    if not verify_columns(df):  # Double check that the columns are there since file versions can change.
        raise ValueError(f'Not all required keys are present in {path}')

    return df


def collate_ztfphot_old(path: PathType, suffix: ZTFPhotSuffix = ZTFPhotSuffix.limits) -> Photometry:
    path = Path(path).resolve(strict=True)
    phot = pd.concat([read_ztfphot_old(p) for p in path.glob(suffix)])
    phot = PhotFactory.from_df(phot)
    return phot


def collate_ztf_detections(path: PathType) -> Photometry:
    """
    Give a path to a directory containing ZTF photometry, return a Photometry object with the detections.
    """
    warnings.warn('collate_ztf_detections is deprecated. Use collate_ztfphot ', DeprecationWarning)
    return collate_ztfphot_old(path, ZTFPhotSuffix.detections)


def collate_ztf_limits(path: PathType) -> Photometry:
    """
    Give a path to a directory containing ZTF photometry, return a Photometry object with the limits.
    """
    warnings.warn('collate_ztf_limits is deprecated. Use collate_ztfphot_limits', DeprecationWarning)
    return collate_ztfphot_old(path, ZTFPhotSuffix.limits)



from typing import Iterator, Optional
import seaborn as sns

DEFAULT_COLORS = {
    "u": "lightsteelblue",
    "g": "forestgreen",
    "r": "tab:red",
    "i": "purple",
    "z": "fuchsia",
    "y": "gold",
    "U": "lightsteelblue",
    "B": "royalblue",
    "V": "olivedrab",
    "R": "darkred",
    "I": "darkviolet",
    "J": "sandybrown",
    "H": "saddlebrown",
    "K": "coral",
}

GENERIC_COLORS = sns.color_palette()

ColorType = str | tuple[float, float, float]
ColorIterable = (
    dict[str, str] | dict[str, ColorType] | list[ColorType] | sns.palettes._ColorPalette 
)


class Color:
    def __init__(self, colors: ColorIterable = DEFAULT_COLORS) -> None:
        self.colors = colors
        self._color_iter = self._set_color_iter()

    def _set_color_iter(self) -> Iterator[ColorType]:
        return iter(
            self.colors.values() if isinstance(self.colors, dict) else self.colors
        )

    def _refresh_color_iter(self) -> None:
        self._color_iter = self._set_color_iter()

    def get_next_color(self) -> ColorType:
        try:
            return next(self._color_iter)
        except StopIteration:
            self._refresh_color_iter()
            return next(self._color_iter)

    def get_color(self, band: Optional[str] = None) -> ColorType:
        if band is None:
            return self.get_next_color()
        if isinstance(self.colors, dict):
            return self.colors.get(band, self.get_next_color())
        return self.get_next_color()



import enum

from pathlib import Path
from types import ModuleType
from typing import Any, Sequence, TypeGuard, cast, Protocol, overload
from dataclasses import Field
from astropy.coordinates.sky_coordinate import SkyCoord
from astropy.units.quantity import Quantity
from astropy.units.core import Unit, dimensionless_unscaled
from numpy.typing import DTypeLike, NDArray
from pandas import Series
import numpy as np
from requests import HTTPError
Number = int | float | Quantity
PathType = Path | str


class SupportsArray(Protocol):
    @overload
    def __array__(self, __dtype: DTypeLike = ...) -> NDArray:
        ...
    @overload
    def __array__(self, dtype: DTypeLike = ...) -> NDArray:
        ...

    @overload
    def __getitem__(self, __key: int) -> Number:
        ...

    @overload
    def __getitem__(self, __key: slice) -> 'SupportsArray':
        ...


QuantityArrayType = Quantity | SupportsArray | np.ndarray[Number, Any]


def _is_quantity_arr(val: Any) -> TypeGuard[Quantity]:
    return isinstance(val[0], Quantity)

def _is_quantity(val: Any) -> TypeGuard[Quantity]:
    return isinstance(val, Quantity)

def quantity_array(array: QuantityArrayType) -> Quantity:
    if isinstance(array, Quantity):
            return array
    if _is_quantity_arr(array):
        unit: Unit = array[0].unit
        return Quantity([val.value for val in array], unit)
    
    return np.array(array) * dimensionless_unscaled


class StrEnum(str, enum.Enum):
    """Enum with string values."""
    def __str__(self) -> str:
        return str(self.value)


def get_field_dtype_default(_field: Field) -> tuple[DTypeLike, float | int | bool | str]:
    dtype = float
    default = np.nan
    if isinstance(_field.default, Series):
        if _field.default.dtype == object:
            dtype = str
            default = ""
        if _field.default.dtype == bool:
            dtype = bool
            default = True
        if _field.default.dtype == int:
            dtype = int
            default = -99
    return dtype, default


class VerbosePrinter:
    """A class to print messages if verbose is True."""
    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    def print(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self.print(*args, **kwargs)

FloatArr = NDArray[np.float64]


def tendrils_api() -> ModuleType:
    """Use tendrils to access FLOWS API if
    optional dependency tendrils is installed.
    """
    try:
        from tendrils import api
    except ImportError:
        raise ImportError("tendrils is not installed. To connect to FLOWS database,"
                          "install tendrils with `pip install tendrils`.")

    return api    

def get_flows_sninfo(snname: str) -> tuple[SkyCoord, float]:
    """Get SN info from FLOWS database."""
    api = tendrils_api()
    if snname.startswith("SN"):
        snname = snname[2:]
    sninfo = api.get_target(snname)
    coords = SkyCoord(ra=sninfo["ra"], dec=sninfo["decl"], unit="deg", frame="icrs")
    redshift = sninfo["redshift"]
    return coords, redshift


def is_http_error(err: Exception) -> TypeGuard[HTTPError]:
    """Check if an exception is an HTTPError."""
    return hasattr(err, "response")



from dataclasses import dataclass
from importlib.machinery import ModuleSpec
from typing import Collection, Hashable, Mapping, MutableMapping, Optional, Any, Iterable, TypeVar, cast
from pathlib import Path
from unittest.util import strclass
import pandas as pd
from svo_filters import svo
import astropy.units as u
import numpy as np
from dust_extinction.parameter_averages import F99
from importlib.util import find_spec
from numpy.typing import NDArray
from specutils import Spectrum1D
from specutils.manipulation import FluxConservingResampler

from .utils import StrEnum, PathType, Number
FLAM = u.erg / u.cm ** 2 / u.s / u.angstrom

@dataclass
class FilterProfile:
    wave: u.Quantity
    throughput: NDArray


class MagSys(StrEnum):
    AB = 'AB'
    Vega = 'Vega'


def wave_pivot(throughput: NDArray, wave: u.Quantity) -> u.Quantity:
    return np.sqrt(np.trapz(throughput) / np.trapz(throughput / wave ** 2))


def get_resampled_vega(waves: u.Quantity) -> u.Quantity:
    vega_data = np.genfromtxt(Path(__file__).parent.absolute()/'vega.dat', unpack=True)[: 2]
    spec = Spectrum1D(spectral_axis=vega_data[0] * u.AA, flux=vega_data[1] * FLAM)
    new_spec = FluxConservingResampler()(spec, waves)
    return new_spec.flux


def wave_eff(waves: u.Quantity, tputs: NDArray,
             magsys: MagSys = MagSys.AB, flux_units: u.Quantity = FLAM) -> u.Quantity:
    """magsys has to be one of AB or Vega
    """
    if magsys == MagSys.Vega:
        jj_spec_flux = get_resampled_vega(waves)
    elif magsys == MagSys.AB:
        refjy = 3631.0 * u.Jy
        jj_spec_flux = np.ones_like(tputs) * refjy.to(flux_units, equivalencies=u.spectral_density(waves))
    else:
        raise ValueError("`magsys` has to be one of `AB` or `Vega`")
    top = np.trapz((waves * tputs * jj_spec_flux), x=waves)
    bot = np.trapz((tputs * jj_spec_flux), x=waves)
    _wave_eff = top / bot
    return _wave_eff.to(u.AA)


class SVOFilter(svo.Filter):
    MagSys: "MagSys"
    flux_units: u.Quantity = FLAM

    def __init__(self, band: str, profile: Optional[FilterProfile] = None,
                 mag_sys: Optional["MagSys"] = None, **kwargs: Any):
        if profile is None:
            super().__init__(band, wave_units=u.AA, **kwargs)
        else:
            self.wave = profile.wave.to(u.AA)
            self.throughput = profile.throughput
            self.MagSys = mag_sys if mag_sys is not None else MagSys.AB
        self._waves = cast(u.Quantity, self.wave.flatten()).to(u.AA)
        self._tputs = self.throughput.flatten()
        self._wave_pivot = wave_pivot(self.throughputs, self.waves)
        self._wave_eff = wave_eff(self.waves, self.throughputs, self.magsys)

    @property
    def magsys(self) -> "MagSys":
        return self.MagSys

    @magsys.setter
    def magsys(self, value: str):
        self.MagSys = MagSys(value)

    @property
    def waves(self) -> u.Quantity:
        return self._waves

    @waves.setter
    def waves(self, value: u.Quantity):
        self._waves = value.to(u.AA)

    @property
    def throughputs(self) -> NDArray:
        return self._tputs

    @property
    def wave_pivot(self) -> u.Quantity:
        return self._wave_pivot

    @wave_pivot.setter
    def wave_pivot(self, value: u.Quantity):
        self._wave_pivot = value.to(u.AA)

    @property
    def wave_eff(self) -> u.Quantity:
        return self._wave_eff

    @wave_eff.setter
    def wave_eff(self, value: u.Quantity):
        self._wave_eff = value.to(u.AA)

    def write_filter(self, path: PathType) -> None:
        tempdf = pd.DataFrame({'#lambda': self.waves.value,
                               'transmission': self.throughputs.flatten()})
        tempdf.to_csv(path, sep=' ', index=False)


FLOWS_FILTERS = {
    "u": SVOFilter("SDSS.u"),
    "g": SVOFilter("PS1.g"),
    "r": SVOFilter("PS1.r"),
    "i": SVOFilter("PS1.i"),
    "z": SVOFilter("PS1.z"),
    "B": SVOFilter('MISC/APASS.B'),
    "V": SVOFilter('MISC/APASS.V'),
    "R": SVOFilter('NOT/ALFOSC.Bes_R'),
    "J": SVOFilter('2MASS.J'),
    "H": SVOFilter('2MASS.H'),
    "K": SVOFilter('2MASS.Ks')
}

_MAG_SYS = {
    "u": 'AB',
    "g": 'AB',
    "r": 'AB',
    "i": 'AB',
    "z": 'AB',
    "B": 'Vega',
    "V": 'Vega',
    "R": 'Vega',
    "J": 'Vega',
    "H": 'Vega',
    "K": 'Vega'
}


for name, filt in FLOWS_FILTERS.items():
    filt.magsys = _MAG_SYS[name]
    FLOWS_FILTERS[name] = filt
    if name == 'B':
        filt.zp = 6.49135e-9 * FLAM
    if name == 'V':
        filt.zp = 3.73384e-9 * FLAM


def get_flows_filter(band: str) -> SVOFilter:
    if band not in FLOWS_FILTERS.keys():
        raise ValueError(f"Band: `{band}` not found in flows filter list: {set(FLOWS_FILTERS.keys())}")
    return FLOWS_FILTERS[band]


def zero_point_flux(svo_filt: SVOFilter, magsys='AB') -> u.Quantity:
    if magsys == 'Vega':
        return svo_filt.zp
    elif magsys != "AB":
        raise ValueError("`magsys` has to be one of `AB` or `Vega`")
    refjy = 3631.0 * u.Jy
    return refjy.to(FLAM, equivalencies=u.spectral_density(svo_filt.wave_pivot))


@dataclass
class Filter:
    """
    A class to represent a filter.
    Stores the filter name, the filter object from the SVO Filter Profile Service,
    the shift and color for plotting,
    as well as the effective wavelength and zero point flux (defaulting to zero),
    if relevant.
    """
    name: str
    wave_eff: u.Quantity = u.Quantity(0, u.AA)
    magsys: MagSys = MagSys.AB
    zp: u.Quantity = u.Quantity(0., FLAM)  # zero point flux in erg/s/cm^2/A
    plot_shift: float = 0
    plot_color: Optional[str | tuple[float, float, float]] = None
    svo: Optional[Any] = None  # should be SVOFilter but that's broken.
    svo_name: Optional[str] = None
    ext: float = 0.0  # extinction in magnitudes in band.

    def __post_init__(self):
        # Not needed if using factory or type hints.
        self.wave_eff = self.wave_eff << u.AA
        if isinstance(self.magsys, str):
            self.magsys = MagSys(self.magsys)
        if isinstance(self.zp, (int, float)):
            self.zp = self.zp << u.erg / u.cm ** 2 / u.s / u.AA
        # enf of defensive unit tests
        self.svo = self.get_svo_filter()
        if self.svo is not None:
            self.wave_eff = self.svo.wave_eff
            self.zp = zero_point_flux(self.svo, self.magsys)
            self.svo_name = self.svo.name

    def get_svo_filter(self) -> Optional[SVOFilter]:
        if isinstance(self.svo, SVOFilter):
            return self.svo
        if self.svo_name is not None:
            try:
                return SVOFilter(self.svo_name)  # will error if you don't know what you are doing.
            except IndexError:
                pass
        try:
            return get_flows_filter(self.name)
        except ValueError:
            pass
        try:
            return SVOFilter(self.name)
        except IndexError:
            return None

    def set_extinction(self, ebv: float, rv: float = 3.1) -> "Filter":
        ext = F99(Rv=rv)
        self.ext = ext(self.wave_eff) * rv * ebv  # type: ignore
        return self
        

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "wave_eff": self.wave_eff.value,
            "magsys": self.magsys.value,
            "zp": self.zp.value,
            "plot_shift": self.plot_shift,
            "plot_color": self.plot_color,
            "svo_name": self.svo_name,
            "ext": self.ext
        }

    @classmethod
    def from_dict(cls, d: MutableMapping[str, Any]) -> "Filter":
        if 'name' not in d:
            raise ValueError("Filter dictionary must have a `name` key.")
        name: str = d.pop('name')
        return cls.filter_factory(name=name, **d)

    def __str__(self):
        return self.name

    @classmethod
    def filter_factory(cls, name: str, **kwargs: Any) -> "Filter":
        """
        A factory method to create a filter from a name.
        """
        wave_eff = kwargs.get("wave_eff", 0.) << u.AA
        zp = kwargs.get("zp", 0.) << u.erg / u.cm ** 2 / u.s / u.AA
        magsys = kwargs.get("magsys", "AB")
        magsys = MagSys(magsys)
        kwargs.update({"name": name, "wave_eff": wave_eff, "zp": zp, "magsys": magsys})

        return cls(**kwargs)

class FilterSorter:
    """
    Use in list.sort(key=FilterSorter(order: list[str]))
    Will use DEFAULT_ORDER if order is None. You can also
    import and append to DEFAULT_ORDER.
    """
    default_order = ["u", "B", "V", "g", "r", "R", "i", "I", "z", "Y", "J", "H", "K"]

    def __init__(self, order: Optional[list] = None):
        self.order = order if order is not None else self.default_order

    def __call__(self, band: str) -> int:
        if band not in self.order:
            return len(self.order)
        return self.order.index(band)

    @staticmethod
    def wave_sort(bands: Iterable[Filter]) -> list[Filter]:
        return sorted(bands, key=lambda x: x.wave_eff.value)

    def order_sort(self, bands: Iterable[Filter]) -> list[Filter]:
        return sorted(bands, key=lambda x: self(x.name))

BandsType = (
    Mapping[str, Filter | Mapping[str, Any]]
    | Collection[str]
    | Collection[Filter]
    | Collection[Mapping[str, Any]]
)

def _get_bands_dict(bands: BandsType) -> dict[str, Filter]:
    # Bands
    new_bands = {}
    if isinstance(bands, MutableMapping):
        for band_name, band in bands.items():
            if isinstance(band, MutableMapping):
                new_bands[band_name] = Filter.from_dict(band)
            if isinstance(band, Filter):
                new_bands[band_name] = band
            else:
                raise ValueError(f"Band must be a mapping or Filter, but is {type(band)}")

    if isinstance(bands, (list, tuple)):
        for band in bands:
            if isinstance(band, str):
                new_bands[band] = Filter(band)
            elif isinstance(band, Filter):
                new_bands[band.name] = band
            elif isinstance(band, MutableMapping):
                new_bands[band["name"]] = Filter.from_dict(band)

    return new_bands


def make_bands(
        bands: BandsType,
        band_order: Optional[list[str]] = None,
        ebv: float = 0.0,
        rv: float = 3.1,
    ) -> Mapping[str, Filter]:
        """
        Creates a sorted dictionary of Filter objects.
        Any unknown filters are added to the end. If band_order is given, the
        order of the bands is set to that, otherwise default_order from
        supernova.filters.FilterSorter is used, which is hard-coded based on wave_eff.
        """
        new_bands = _get_bands_dict(bands)

        wave_sort = band_order is None
        sorter = FilterSorter(band_order)
        if wave_sort:
            new_bands = sorter.wave_sort(new_bands.values())
        else:
            new_bands = sorter.order_sort(new_bands.values())

        return {
            b.name: b 
            if b.svo is None 
            else b.set_extinction(ebv, rv)
            for b in new_bands
        }

def pisco_filter_path() -> Path:
    pisc_spec: ModuleSpec | None = find_spec("piscola")
    if pisc_spec is None: 
        raise ImportError("piscola is not installed. Please install it with `pip install piscola`")
    pisc_folder = pisc_spec.origin
    if pisc_folder is None:
        raise ImportError("Cannot determine install location of Piscola. Got None.")
    return Path(pisc_folder).parent.absolute() / 'filters/'    

PISCO_FILTER_PATH = pisco_filter_path()



import importlib
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Mapping, Optional, Protocol, cast

import astropy.units as u
import numpy as np
import pandas as pd
import piscola
from numpy.typing import NDArray
from numpy._typing import _ArrayLike as ArrayLike

from .filters import FLAM, Filter, MagSys, SVOFilter, zero_point_flux
from .photometry import FluxPhotometry, MagPhotometry, Phot, PhotFactory, Photometry
from .sites import Sites
from .supernova import SN, SNSerializer
from .utils import Number, QuantityArrayType, quantity_array


class GPInterp(Protocol):
    def __init__(self, sn: SN) -> None:
        ...

    def __call__(self) -> pd.DataFrame:
        ...


class PiscolaInterp:
    piscola_names = [
        "time",
        "flux",
        "flux_err",
        "mag",
        "mag_err",
        "zp",
        "band",
        "mag_sys",
    ]

    def __init__(self, sn: SN) -> None:
        self.sn = sn
        self.sns = SNSerializer(self.sn)

    def prepare_piscola_filters(self) -> None:
        """Prepare Piscola filters for interpolation."""
        self.sns.save_piscola_filters(sites=False)
        importlib.reload(piscola)

    @staticmethod
    def _fix_zp(lcdf: pd.DataFrame, correct_zp: float) -> pd.DataFrame:
        """Fix the zero point of the light curves."""
        lcdf["flux"] = piscola.utils.change_zp(lcdf.flux, lcdf.zp, correct_zp)
        lcdf["flux_err"] = piscola.utils.change_zp(lcdf.flux_err, lcdf.zp, correct_zp)
        lcdf["zp"] = correct_zp
        return lcdf

    def make_df(self, lc: Any, init_lcs: Any) -> pd.DataFrame:
        """Make a DataFrame from a Piscola light curve."""
        dfs = []
        for band in lc.bands:
            df = pd.DataFrame(
                {colname: lc[band][colname] for colname in self.piscola_names}
            )
            zp_corr = init_lcs[band].zp
            df = self._fix_zp(df, zp_corr)
            dfs.append(df)
        return pd.concat(dfs)

    def gp_interp(self) -> pd.DataFrame:
        with TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            piscofile = self.sns.make_piscola_file(tmpdir)
            # lcs_df = pd.read_csv(piscofile, delim_whitespace=True, skiprows=2)

            sn = piscola.call_sn(str(piscofile.absolute()))
            sn._fit_lcs()
            lcs = self.make_df(sn._init_lc_fits, sn.init_lcs)
        return lcs

    @staticmethod
    def _fix_site_band_names(df: pd.DataFrame) -> pd.DataFrame:
        """Fix band names."""
        df["site"] = df["band"].str.split("_").str[0]
        df["band"] = df["band"].str.split("_").str[1]
        return df

    def __call__(self) -> pd.DataFrame:
        self.prepare_piscola_filters()
        df = self.gp_interp()
        df = self._fix_site_band_names(df)
        return df.rename(columns={"time": "jd"})


class LCInterp:
    def __init__(
        self, reference_times: ArrayLike, phot: Photometry, band: Filter
    ) -> None:
        self._time = phot.jd
        self.reference_times = reference_times
        self.band = band
        self.svo_filter = band.svo
        self._phot = phot

        if not isinstance(phot, MagPhotometry):
            raise TypeError("Photometry must have `mag` and `mag_err`.")
        self._mag = self.interp(phot.mag)
        self._mag_err = self.interp(phot.mag_err)

        if isinstance(phot, FluxPhotometry):
            self._flux = self.interp(phot.flux)
            self._flux_err = self.interp(phot.flux_err)
        else:
            if self.svo_filter is None:
                raise ValueError(
                    "Photometry must have `flux` and `flux_err` if `svo_filter` is not set."
                )
            self._flux = mag_to_flux(self.mag, self.svo_filter, magsys=self.band.magsys)
            self._flux_err = mag_to_flux_err(self._mag_err, self.flux)

    def interp(self, yp: QuantityArrayType) -> np.ndarray[Number, Any]:
        return np.interp(self.reference_times, self._time, yp)

    @property
    def mag(self):
        return self._mag

    @property
    def mag_err(self):
        return self._mag_err

    @property
    def flux(self):
        return self._flux

    @property
    def flux_err(self):
        return self._flux_err

    def to_phot(self) -> Photometry:
        df = pd.DataFrame(
            {
                "mag": self.mag,
                "mag_err": self.mag_err,
                "flux": self.flux,
                "flux_err": self.flux_err,
                "jd": self.reference_times,
            }
        )
        df["site"] = self._phot.site.unique()[0]
        df["band"] = self.band.name
        df["sub"] = True
        return PhotFactory.from_df(df)


class SNInterp:
    def __init__(
        self,
        sn: SN,
        filters_dict: Mapping[str, Filter],
        reference_times: NDArray[np.float64],
        interp_type: type[GPInterp] = PiscolaInterp,
    ) -> None:
        self.filters_dict = filters_dict
        self.reference_times = reference_times
        self.sn = sn
        self.interp: GPInterp = interp_type(sn)
        self._sitemap = self.sn.sites.sites

    def update_sitemap(
        self, phot: Photometry | pd.DataFrame
    ) -> Photometry | pd.DataFrame:
        new_sites = phot.site.unique()
        sites = Sites(sites=self._sitemap)
        for site in new_sites:
            if site not in sites:
                new_site = sites.add_site(name=site)
                self._sitemap[new_site.id] = new_site
        phot.site = cast(pd.Series, phot.site.apply(lambda x: sites[x].id))
        sites = Sites(sites=self._sitemap.copy())
        for site in self._sitemap.values():
            if site.name not in new_sites:
                sites.remove_site(site)
        self._sitemap = sites.sites
        return phot

    def _fix_df_for_phot(self, df: pd.DataFrame) -> pd.DataFrame:
        df["sub"] = True
        p0 = self.sn.phase_zero if self.sn.phase_zero is not None else df["jd"].iloc[0]
        df["phase"] = df['jd'] - p0
        return df

    def gp_interp(self) -> Photometry:
        df = self.interp()
        df = self._fix_df_for_phot(df)
        df = self.update_sitemap(df)
        if isinstance(df, pd.DataFrame):
            return PhotFactory.from_df(df)
        return df

    def lc_interp(self, phot: Optional[Photometry] = None) -> Photometry:
        if phot is None:
            phot = self.sn.phot
        fac = PhotFactory(Phot())
        for band_name, band in self.filters_dict.items():
            band_phot = phot.masked(phot.band == band_name)
            lc = LCInterp(self.reference_times, band_phot, band)
            fac.concat_phot(lc.to_phot())
        return fac.sn_phot

    def create_interp_sn(self) -> SN:
        interp_phot = self.gp_interp()
        interp_phot_ref = self.lc_interp(interp_phot)
        sn = self.sn
        return SN.from_phot(
            interp_phot_ref,
            sn.name,
            sn.redshift,
            sn.phase_zero,
            sn.sub_only,
            sninfo=sn.sninfo,
            lims=sn.limits,
            sites=Sites(sites=self._sitemap),
        )


def mag_to_flux(
    mag: QuantityArrayType, filt: SVOFilter, magsys=MagSys.AB
) -> u.Quantity:
    qmag = quantity_array(mag)
    if magsys == "AB":
        qmag = qmag << u.ABmag
        return qmag.to(FLAM, equivalencies=u.spectral_density(filt.wave_eff))
    elif magsys != "Vega":
        raise ValueError("`magsys` has to be one of `AB` or `Vega`")

    return 10 ** (qmag.value / -2.5) * zero_point_flux(filt, magsys)


def mag_to_flux_err(mag_err: QuantityArrayType, flux: QuantityArrayType) -> u.Quantity:
    return 2.303 * quantity_array(mag_err) * quantity_array(flux)



from typing import Protocol, Type, Callable, Optional
from pathlib import Path
import pandas as pd

from .converters import Converter, GenericCSVConverter, GenericECSVConverter
from .readers import (PathType, resolve_path, PhotReader, read_pandas_csv,
                      read_astropy_table_as_df_with_times_as_jd)
from .utils import NEEDED_KEYS
from astrosn.photometry import Photometry, PhotFactory

CollatorType = Callable[[PathType], Photometry]


class AbstractCollator(Protocol):
    path: PathType = 'Unset'  # last processed path.
    converter: Type[Converter]
    reader: PhotReader
    processed_files: list[PathType]  # all processed files.
    ignore: bool = True

    def __init__(self, converter: Type[Converter],
                 reader: PhotReader, ignore_processed: bool = True) -> None:
        raise NotImplementedError("This is an abstract Protocol class.")

    def __call__(self, path: PathType) -> Photometry:
        ...

    def process(self, file: PathType) -> pd.DataFrame:
        ...

    def collate(self, path: PathType) -> Photometry:
        ...


class Collator(AbstractCollator):

    def __init__(self, converter: Type[Converter],
                 reader: PhotReader, ignore_processed: bool = True) -> None:
        self.path = "Unset"
        self.converter = converter
        self.reader = reader
        self.processed_files: list[PathType] = []
        self.ignore = ignore_processed
        self.last_used_converter: Optional[Converter] = None

    def __call__(self, path: PathType) -> Photometry:
        self.path = resolve_path(path)
        return self.collate(self.path)

    def process(self, file: PathType) -> pd.DataFrame:
        if self.ignore and file in self.processed_files:
            return pd.DataFrame()  # empty dataframes in pd.concat are ignored.
        self.processed_files.append(file)
        return self.reader(file)

    def collate(self, path: PathType) -> Photometry:
        files = list(Path(path).glob(self.converter.glob_str))
        if len(files) == 0:
            raise FileNotFoundError(f"No files found in {path} with glob {self.converter.glob_str}")
        df = pd.concat([self.process(p) for p in files], ignore_index=True)
        self.last_used_converter = self.converter(df=df)
        phot_df = self.last_used_converter.convert()
        return PhotFactory.from_df(phot_df[NEEDED_KEYS])


collate_ecsv = Collator(GenericECSVConverter, read_astropy_table_as_df_with_times_as_jd)
collate_csv = Collator(GenericCSVConverter, read_pandas_csv)
collate_json = Collator(GenericECSVConverter, pd.read_json)




from typing import Any, cast
from . import plot_lc as plot
from astrosn import SN, Photometry
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.ticker import MultipleLocator


AbsMagPlotParams: plot.PlotParameters = plot.PlotParameters(
    absmag=True, split_by_site=True, label_sites=False, mark_unsub=False
)
AbsMagAxisParams: plot.AxisParameters = plot.AxisParameters(
    xlabel="Restframe days",
    ylabel="Absolute Magnitude",
    x_tick_interval=10,
    invert=True,
    y_tick_interval=1,
    legend_kwargs={"loc": "upper right", "ncol": 3},
)

def _get_x_tick_interval(phot: Photometry) -> float:
    """
    Get the x_tick_interval based on the range of the photometry.
    """
    p_range: float = phot.phase.max() - phot.phase.min()
    return max(p_range // 12, 10)



def plot_flows_absmag(
    sn: SN,
    axpars: plot.AxisParameters = AbsMagAxisParams,
    plotpars: plot.PlotParameters = AbsMagPlotParams,
    secondary_y: bool = True,
    shift: bool = True,
    **fig_kwargs: Any,
) -> tuple[Figure, Axes]:
    """
    Plot the absolute magnitude of a FLOWS supernova.
    """
    sn = sn.restframe()  # Make sure we're in the restframe

    # Check if we can split by site.
    if plotpars.split_by_site:
        try:
            sn.sites.update_markers(plot.get_site_markers(sn.sites.markers))
        except ValueError:
            plotpars.split_by_site = False
            print("Too many sites to plot with unique markers and some site markers were `None`. Plotting all together."
            "If you want to split by site, set the marker of each site to a matplotlib marker string manually."
            "You can use `supernova.plotting.plot._get_markers_to_use` to get a list of unique markers."
            "use `SN.sites.update_markers` to set the markers of each site at once.")

    # Automatically set the x_tick_interval based on the range of the photometry.
    if axpars.x_tick_interval in (10, 10., None):
        _phot = sn.phot
        if sn.sub_only:
            _phot = _phot.masked(_phot.sub.tolist())
        axpars.x_tick_interval = _get_x_tick_interval(_phot)

    # Turn off autoshifts if desired.
    if not shift:
        plotpars.shifts = [0. for b in sn.bands]

    # plot
    plotter = plot.Plotter(sn, plot.plot_abs_mag, plotpars, axpars)
    fig, ax = plotter(ax=None, **fig_kwargs)

    # Make tight layout
    fig = cast(Figure, fig)
    fig.tight_layout()

    # Add app mag as secondary
    if secondary_y:
        secax: Axes = cast(Axes, ax.secondary_yaxis('right', functions=(lambda l: l+sn.distance, lambda x: x-sn.distance)))
        secax.set_ylabel('App. Mag', fontsize=16, fontweight='semibold')
        minor = 1/axpars.y_minors if axpars.y_minors else 0.2
        secax.yaxis.set_minor_locator(MultipleLocator(minor)) 

    return fig, ax



from dataclasses import dataclass, asdict, field, fields
from typing import Any, MutableMapping, TypeVar, Optional, Collection, cast
import warnings

import numpy as np
import pandas as pd
from typing_extensions import Unpack, TypedDict, NotRequired

from .photometry import Photometry
from .utils import tendrils_api

SiteType = TypeVar('SiteType', int, str)
SiteDict = MutableMapping[int, str]


@dataclass
class Site:
    id: int
    name: str
    marker: Optional[str] = None
    err_scale: float = 1.  # some photometry pipelines misrepresent errors.

    def __call__(self):
        return self.name

SiteKeys = TypedDict("SiteKeys", {'id': int, 'name': str, 'marker': NotRequired[str],
                     'err_scale': NotRequired[float]})
SiteKeysGenID = TypedDict("SiteKeys", {'id': NotRequired[int], 'name': str, 'marker': NotRequired[str],
                     'err_scale': NotRequired[float]})


@dataclass
class Sites:
    sites: dict[int, Site] = field(default_factory=dict)
    site_names: list[str] = field(default_factory=list, init=False)
    site_ids: list[int] = field(default_factory=list, init=False)
    sites_by_name: dict[str, Site] = field(default_factory=dict, init=False)
    markers: dict[int, Optional[str]] = field(default_factory=dict)
    site_err_scales: dict[int, float] = field(default_factory=dict)
    # for repeatable siteid generation:
    rng: np.random.Generator = field(default_factory=np.random.default_rng)

    def __post_init__(self):
        self._refresh_sites()

    def __getitem__(self, site: SiteType) -> Site:
        if site in self.sites:  # int
            return self.sites[site]
        elif site in self.sites_by_name:  # str
            return self.sites_by_name[site]
        raise KeyError(f"site not found: {site}")

    def __contains__(self, site: SiteType) -> bool:
        return site in self.site_names or site in self.site_ids

    def __iter__(self):
        yield from self.sites.values()

    def _refresh_sites(self):
        self.site_names = [site.name for site in self.sites.values()]
        self.site_ids = [site.id for site in self.sites.values()]
        self.sites_by_name = {value.name: value for value in self.sites.values()}
        self.markers = {site.id: site.marker for site in self.sites.values()}

    def update_markers(self, site_markers: Optional[dict[int, str]] = None) -> None:
        if site_markers is not None:
            self.markers = site_markers  # type: ignore
        for site in self.sites.values():
            site.marker = self.markers[site.id]

    def get_marker(self, site: SiteType) -> str | None:
        return self[site].marker

    def get_err_scale(self, site: SiteType) -> float:
        return self[site].err_scale

    def asdict(self) -> dict:
        return asdict(self)

    def to_df(self) -> pd.DataFrame:
        return pd.DataFrame(self.asdict()['sites']).T

    def __len__(self) -> int:
        return len(self.site_ids)

    def generate_id(self):
        return self.rng.choice(list(set(range(100)) - set(self.site_ids)))

    def add_site(self, **site_kwargs: Unpack[SiteKeysGenID]) -> Site:
        if site_kwargs.get('id') is None:
            site_kwargs['id'] = self.generate_id()
        siteid = site_kwargs.get('id', self.generate_id())
        self.sites[siteid] = Site(**site_kwargs)
        self._refresh_sites()
        return self.sites[siteid]

    def remove_site(self, site: Site) -> Site:
        rmd_site = self.sites.pop(site.id)
        self._refresh_sites()
        return rmd_site

    @classmethod
    def from_sitemap(cls, sitemap: SiteDict) -> 'Sites':
        return cls(sites={k: Site(id=k, name=v) for k, v in sitemap.items()})

    @classmethod
    def from_dict(cls, d: dict[int, SiteKeys]) -> 'Sites':
        site_names = [f.name for f in fields(Site)]
        required_keys = [f for f in fields(Site) if not isinstance(f.default, f.type)]

        return cls(
            sites={k: Site(**{sk: sv for sk, sv in v.items() if sk in site_names}) # type: ignore
                   for k, v in d.items() if all([vk for vk in v if vk in required_keys])}
        )

    @classmethod
    def from_phot(cls, phot: pd.DataFrame | Photometry) -> 'Sites':
        return cls.from_list(dict(phot.site.unique().astype(str)))

    @classmethod
    def from_list(cls, sitenames: Collection[str]) -> 'Sites':
        return cls(sites={i: Site(i, s) for i, s in enumerate(sitenames)})

    @classmethod
    def from_df(cls, df: pd.DataFrame | pd.Series) -> 'Sites':
        if isinstance(df, pd.Series):
            return cls.from_sitemap(cast(SiteDict, df.to_dict()))
        if 'id' not in df:
            df['id'] = df.index
        if 'name' not in df:
            df.rename({df.columns[0]: 'name'}, axis=1, inplace=True)
        return cls.from_dict(cast(dict[int, SiteKeys], df.to_dict('index')))


def get_flows_sites() -> SiteDict:
    flows_sites = {8: 'LT', 5: 'NOT', 1: 'LCOGT', 0: 'ZTF'}
    try:
        api = tendrils_api()
        sites = api.get_all_sites()
        flows_sites = cast(SiteDict, {s['siteid']:s['sitename'] for s in sites})
        flows_sites[1] = 'LCOGT'
        flows_sites[0] = 'ZTF'
    except ImportError:
        warnings.warn("tendrils_api not found, using default flows_sites")
    
    return flows_sites
    

flows_sites = get_flows_sites()
site_markers: SiteDict = {1: 's', 5: 'd', 8: 'v', 0: 'o'}
site_err_scales: dict[int, int] = {1: 2, 5: 2, 8: 5, 0: 1}



from astrosn import Photometry
from typing import get_args
import pandas as pd
from dataclasses import fields

from astrosn.photometry import BBLumPhot, FluxPhot, LumPhot, MagPhot, Phot, PhotFactory


phot_classes = [MagPhot, FluxPhot, LumPhot, Phot, BBLumPhot]

def test_phot_load():
    df = pd.read_csv('inputs/SN2020lao_phot.csv', index_col=0).squeeze("columns")
    phot = PhotFactory.from_df(df)
    assert isinstance(phot, Photometry)

def test_phot_partial_init():
    jds = pd.Series([1., 2., 3.])
    bands = pd.Series(['g', 'r', 'i'])

    for phot_class in phot_classes:
        phot = phot_class(jd=jds, band=bands)
        for field in fields(phot):
            field_instance = getattr(phot, field.name)
            assert len(getattr(phot, field.name)) == len(phot) == 3
            assert field.default.dtype == field_instance.dtype
    #for phot_class in get_args(Photometry):
        
        
        # phot = phot_class(jd=jds, band=bands)
        # for field in fields(phot):
        #     field_instance = getattr(phot, field.name)
        #     assert len(getattr(phot, field_instance)) == len(phot) == 3
        #     assert field.default.dtype == field_instance.dtype


if __name__ == '__main__':
    test_phot_load()
    test_phot_partial_init()



"""
Supernova analysis library by @emirkmo
"""
from .version import version as __version__
from .version import version_info as __version_info__
from .supernova import *
from .interp import *
from .filters import *
from .photometry import *
from .sites import Sites, Site, flows_sites



from dataclasses import dataclass, field, asdict, replace
from enum import Enum
import warnings
from typing import Iterable, Optional, Any, Callable, Sequence, TypeGuard, cast
import corner
from lmfit.minimizer import MinimizerResult
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from matplotlib.ticker import MultipleLocator, AutoMinorLocator
from matplotlib.figure import Figure
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from matplotlib.legend import Legend
from matplotlib.axes import Axes
from matplotlib.container import ErrorbarContainer
from matplotlib.collections import LineCollection
import inspect

from astrosn.filters import Filter
from astrosn.sites import Site
from astrosn.supernova import SN
from astrosn.photometry import (
    HasFlux,
    HasLuminosity,
    HasMag,
    ImplementsFlux,
    ImplementsLuminosity,
    ImplementsMag,
    MagPhotometry,
    Photometry,
    Series,
)
from .colors import ColorIterable, DEFAULT_COLORS, ColorType, Color


# Plot styling reasonable defaults.
sns.set_style("ticks")
sns.set_context("paper", font_scale=2)
mpl.rcParams["axes.labelweight"] = "semibold"
plot_defaults = {
    "label": "",
    "alpha": 1.0,
    "marker": "o",
    "markersize": 5,
    "linestyle": "None",
    "color": None,
}
snake_defaults = {"alpha": 0.7, "zorder": 0}
figure_defaults = {
    "figsize": (8, 6),
    "dpi": 200,
}
legend_defaults = {
    "frameon": True,
    "columnspacing": 0.3,
    "handletextpad": 0.1,
    "handlelength": 1.5,
}
default_color = Color(DEFAULT_COLORS)
PlotType = Callable[..., Axes]


def plot_snake(
    x: Series[float], y: Series[float], z: Series[float], ax: Axes, **plot_kwargs: Any
) -> Axes:
    kwargs = {**snake_defaults, **plot_kwargs}
    ax.fill_between(x, y - z, y + z, **kwargs)  # type: ignore
    return ax


def plot_mag(
    phot: ImplementsMag,
    ax: Axes,
    shift: float = 0,
    error_snake: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    if error_snake:
        ps = phot.mag + shift
        return plot_snake(phot.phase, phot.mag + shift, phot.mag_err, ax, **plot_kwargs)
    kwargs = {**plot_defaults, **plot_kwargs}
    ax.errorbar(phot.phase, phot.mag + shift, phot.mag_err, **kwargs)
    return ax


def plot_flux(
    phot: ImplementsFlux,
    ax: Axes,
    shift: float = 0,
    error_snake: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    if error_snake:
        return plot_snake(
            phot.phase, phot.flux + shift, phot.flux_err, ax, **plot_kwargs
        )
    kwargs = {**plot_defaults, **plot_kwargs}
    ax.errorbar(phot.phase, phot.flux + shift, phot.flux_err, **kwargs)
    return ax


def plot_lum(phot: ImplementsLuminosity, ax: Axes, **plot_kwargs: Any) -> Axes:
    kwargs = {**plot_defaults, **plot_kwargs}
    ax.errorbar(phot.phase, phot.lum, phot.lum_err, **kwargs)
    return ax


def plot_lim(
    phot: Photometry,
    ax: Axes,
    shift: float = 0,
    as_flux: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    lim_defaults = {"label": "", "marker": "v", "markerfacecolor": "None"}
    kwargs = {**plot_defaults, **lim_defaults, **plot_kwargs}
    if as_flux:
        if isinstance(phot, HasFlux):
            ax.plot(phot.phase, phot.flux + shift, **kwargs)
    else:
        if not isinstance(phot, HasMag):
            raise TypeError(
                "Cannot plot limits as magnitudes if photometry does not have `mag`."
            )
        ax.plot(phot.phase, phot.mag + shift, **kwargs)
    return ax


def plot_lims(
    sn: SN,
    band: Filter,
    ax: Axes,
    absmag: bool = False,
    as_flux: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    """
    Plot limits for a given band.
    """
    phot = sn.band(band.name, return_absmag=absmag, lims=True)
    ax = plot_lim(phot, ax, band.plot_shift, as_flux, **plot_kwargs)
    return ax


def plot_band(
    sn: SN,
    band: Filter,
    ax: Axes,
    site: str = "all",
    plot_type: PlotType = plot_mag,
    absmag: bool = False,
    mark_unsub: bool = False,
    **kwargs: Any,
) -> Axes:
    """
    Plot a single band.
    """
    _site = sn.sites[site] if site != "all" else site
    marker = kwargs.get("marker", get_marker(_site))

    band_phot = sn.band(band.name, site=site, return_absmag=absmag)
    ax = plot_type(band_phot, ax, **kwargs)
    if not mark_unsub or sn.sub_only:
        return ax

    # Put an X tick mark on the plot for unsubtracted photometry.
    sn.sub_only = True
    sub_phot = sn.band(band.name, site=site, return_absmag=absmag)
    sn.sub_only = False
    unsub_phot = sn.band(band.name, site=site, return_absmag=absmag)
    # Get ONLY unsubtracted photometry:
    conds = [not any(np.isclose(jd, sub_phot.jd, atol=1e-7)) for jd in unsub_phot.jd]
    unsub_phot = unsub_phot.masked(conds)

    unsub_marker = "x"
    newkwargs = {
        **kwargs,
        **{"marker": unsub_marker, "color": "black", "linestyle": "None", "label": ""},
    }
    newkwargs["markersize"] = newkwargs.get(
        "markersize", plot_defaults["markersize"] + 2
    )
    newkwargs["zorder"] = 5
    ax = plot_type(unsub_phot, ax, **newkwargs)
    return ax


def _is_single_figure(axs: list[list[Axes]] | Axes) -> TypeGuard[Axes]:
    return not isinstance(axs, list)


def _is_multi_figure(axs: list[list[Axes]] | Axes) -> TypeGuard[list[list[Axes]]]:
    return isinstance(axs, list)


def make_figure(**fig_kwargs: Any) -> tuple[Figure, list[list[Axes]] | Axes]:
    fig_kwargs = figure_defaults | fig_kwargs
    fig, axs = plt.subplots(**fig_kwargs)
    if _is_single_figure(axs):
        return fig, axs
    return fig, axs


def make_single_figure(**fig_kwargs: Any) -> tuple[Figure, Axes]:
    fig, axs = make_figure(**fig_kwargs)
    if _is_single_figure(axs):
        return fig, axs
    elif _is_multi_figure(axs):
        ax = cast(Axes, axs[0][0])
        return fig, ax

    raise TypeError("Unknown axes type.")


def make_plot_shifts(
    sn: SN,
    shifts: Optional[Sequence[float | int]] = None,
    reversed_for_abs: bool = False,
) -> SN:
    """
    Modify state of SN object with the plot shift
    for each filter set. If Shifts are not given
    then they will be +/- 1 for each filter.
    By default, first shift is negative, middle is zero, and last is positive.
    If reversed_for_abs is True, then the first shift is positive, middle is zero, and last is negative
    for use in plotting in absolute magnitude space. In that case, input shifts order is also reversed.
    """
    n_bands = len(sn.bands)
    if shifts is not None and len(shifts) < n_bands:
        raise ValueError(
            "shifts must be the same or greater "
            "length as the number of bands or `None`."
        )

    if shifts is None:
        use_shifts = np.linspace(0, n_bands, n_bands + 1) - (n_bands // 2)
    else:
        use_shifts = shifts

    if reversed_for_abs:
        use_shifts = use_shifts[::-1]

    for i, f in enumerate(sn.bands.values()):
        f.plot_shift = use_shifts[i]

    return sn


def make_plot_colors(sn: SN, colors: Optional[ColorIterable] = None) -> SN:
    """
    Modify state of SN object with the plot color
    for each filter set. If colors are not given
    then they will be a default set of colors.
    colors should be a dict or list of matplotlib
    compatible colors (ex.:3 floats or a color string).
    """
    if colors is None:
        use_colors = sns.color_palette("Paired", n_colors=len(sn.bands))
    else:
        use_colors = colors
    # if isinstance(use_colors, sns.colors._ColorPalette.LienarSeg
    if isinstance(use_colors, LinearSegmentedColormap | ListedColormap):
        raise TypeError(
            "Color map LinearSegmentedColormap | ListedColormap is not a valid color iterable."
        )

    if len(use_colors) < len(sn.bands):
        raise ValueError(
            "colors must be the same or greater "
            "length as the number of bands or `None`."
        )
    if isinstance(use_colors, list):
        use_colors = {
            band.name: color for band, color in zip(sn.bands.values(), use_colors)
        }
    for band in sn.bands.values():
        band.plot_color = use_colors[band.name]
    return sn


def _get_markers_to_use() -> list[str]:
    """
    Get a list of markers to use for plotting.
    """
    markers = ["o", "d", "s"]
    # Pad smartly with markers using more legible markers first.
    filled_markers = list(Line2D.filled_markers)
    # remove markers that are already in use
    for m in markers:
        filled_markers.remove(m)

    # pad filled_markers with unfilled markers
    markers += filled_markers
    markers += ["1", "2", "3", "4", "+", "X"]
    return markers


def get_site_markers(site_markers: dict[int, Optional[str]]) -> dict[int, str]:
    """
    Given a dictionary of site markers: mapping of int to str or None
    return dictionary of int to str using more legible markers first.
    """
    markers = _get_markers_to_use()
    new_site_markers = {}
    if len(site_markers) > len(markers):
        raise ValueError("Too many sites to plot with unique markers.")
    used_markers = set(site_markers.values())
    if None not in used_markers:
        # unnecessary, just for pylance:
        new_site_markers = {k: v for k, v in site_markers.items() if v is not None}
        return new_site_markers
    used_markers.remove(None)
    for m in used_markers:
        if m in markers:
            markers.remove(m)

    for site, marker in site_markers.items():
        if marker is None:
            new_site_markers[site] = markers.pop(0)
        else:
            new_site_markers[site] = marker

    return new_site_markers


def get_label(filt: Filter, site_name: Optional[str] = "") -> str:
    """
    Get a label for a filter, optionally with a site name.
    If sitename is not desired, pass an empty string or None
    """
    site_name = "" if site_name is None else site_name
    if filt.plot_shift == 0:
        label = f"${filt.name}$ {site_name}"
        return label.strip()
    shift_label = f"{filt.plot_shift:.1f}"
    if not shift_label.startswith("-") and shift_label != "":
        shift_label = f"+{shift_label}"
    if shift_label.endswith(".0"):
        shift_label = shift_label[:-2]
    label = f"${filt.name}${shift_label} {site_name}"
    return label.strip()


def get_marker(site: Site | str) -> str:
    """
    Get a marker for a site.
    """
    default_marker = _get_markers_to_use()[0]
    if site == "all":
        return default_marker
    if isinstance(site, str):
        raise TypeError("site must be a Site object or 'all'")
    marker = site.marker
    if marker is None:
        return default_marker
    return marker


def get_color(band: Filter, colors: Color = default_color) -> ColorType:
    color = band.plot_color
    if color is not None:
        return color
    return colors.get_color(band.name)


def get_plot_kwargs(
    site: str | Site,
    band: Filter,
    site_label: str = "",
    update_label: bool = True,
    **plot_kwargs: Any,
) -> dict:
    if plot_kwargs is None:
        plot_kwargs = {}
    # if plot_kwargs.get('marker', None) == None:
    plot_kwargs["marker"] = get_marker(site)
    # if plot_kwargs.get('color', None) is None:
    plot_kwargs["color"] = get_color(band)
    if update_label:
        plot_kwargs["label"] = get_label(band, site_label)
    return {**plot_defaults, **plot_kwargs}


def plot_all_bands(
    sn: SN,
    ax: Axes,
    site: str = "all",
    plot_type: PlotType = plot_mag,
    absmag: bool = False,
    make_labels: bool = True,
    mark_unsub: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    # _ax: Axes, _sn: SN, _site: str, _site_label: str, _marker: str, _plot_kwargs: dict,
    #           update_label: bool = True
    """
    Plot all bands of a site, or of all sites if not given, on a single axis.
    """
    _site = "all"
    if site != "all":
        _site = sn.sites[site]

    for band in sn.bands.values():
        plot_kwargs = get_plot_kwargs(
            _site, band, update_label=make_labels, **plot_kwargs
        )
        if not make_labels:
            plot_kwargs["label"] = None
        try:
            band_phot = sn.band(
                band.name,
                site=site,
                return_absmag=absmag,
                flux=plot_type == plot_flux,
                lims=plot_type == plot_lims,
            )
        except KeyError as e:
            warnings.warn(
                f"Could not get photometry for {band.name} in {sn.name},\n got exception: {e}"
            )
            band_phot = None

        if band_phot is not None and len(band_phot) > 0:
            ax = plot_band(
                sn=sn,
                band=band,
                ax=ax,
                site=site,
                plot_type=plot_type,
                absmag=absmag,
                shift=band.plot_shift,
                mark_unsub=mark_unsub,
                **plot_kwargs,
            )
    return ax


def get_unsub_mark(
    markersize: int = 7, label: str = "Unsub"
) -> tuple[list[Line2D], list[str]]:
    handles = [
        Line2D(
            [0], [0], color="black", marker="x", linestyle="None", markersize=markersize
        )
    ]
    labels = [label]
    return handles, labels


def get_photcol(phot: Photometry, plot_type: PlotType) -> Series[float]:
    if plot_type == plot_mag:
        if isinstance(phot, HasMag):
            return phot.mag
    if plot_type == plot_flux:
        if isinstance(phot, HasFlux):
            return phot.flux
    if plot_type == plot_lum:
        if isinstance(phot, HasLuminosity):
            return phot.lum
    if plot_type == plot_lims:
        if isinstance(phot, HasFlux):
            return phot.flux
        elif isinstance(phot, HasMag):
            return phot.mag

    raise ValueError(f"Invalid plot type: {plot_type}")



def plot_split_by_site(
    sn: SN,
    ax: Axes,
    plot_type: PlotType = plot_mag,
    absmag: bool = False,
    label_sites: bool = True,
    mark_unsub: bool = False,
    **plot_kwargs: Any,
) -> Axes:
    """
    Plot all bands of a supernova, split by site.
    """
    sn.sites.update_markers(get_site_markers(sn.sites.markers))
    #first = True
    legend_labels = []
    legend_handles = []
    for site in sn.sites:
        make_labels = False
        ax = plot_all_bands(
            sn,
            ax,
            site=site.name,
            plot_type=plot_type,
            absmag=absmag,
            make_labels=make_labels,
            mark_unsub=mark_unsub,
            **plot_kwargs,
        )
        #first = False
        if label_sites: # add site as black markers
            legend_labels.append(site.name)
            usemarker = site.marker if site.marker is not None else _get_markers_to_use()[0]
            legend_handles.append(
                Line2D(
                    [0],
                    [0],
                    color="black",
                    marker=usemarker,
                    linestyle="None",
                    markersize=plot_kwargs.get("markersize", 6),
                )
            )

    for band in sn.bands:
        phot = sn.band(band, flux=plot_type == plot_flux,
                lims=plot_type == plot_lims)
        phot_col = get_photcol(phot, plot_type=plot_type)
        if len(phot_col.dropna()) > 0:


    
            temp_kwargs = get_plot_kwargs('all', sn.bands[band], **plot_kwargs)
            legend_labels.append(temp_kwargs['label'])

            line = Line2D(
                    [0],
                    [0],
                    color=temp_kwargs.get('color', sn.bands[band].plot_color),
                    marker=_get_markers_to_use()[0],
                    linestyle=temp_kwargs.get('linestyle', 'None'),
                    markersize=temp_kwargs.get("markersize", 6),
                )
            # barline = LineCollection(np.empty((2,2,2)))
            # err = ErrorbarContainer((line, [line], [barline]), has_xerr=True, has_yerr=True)
            legend_handles.append(line)

    leg, ax = _safe_draw_legend(ax, legend_handles, legend_labels)
    return ax


def update_legend_handles_labels(
    ax: Axes, legend_handles: list[Any], legend_labels: list[str]
) -> tuple[Axes, list[Any], list[str]]:
    handles, labels = _safe_get_legend_handles_labels(ax)
    for handle, label in zip(handles, labels):
        if label in legend_labels:
            continue
        legend_handles.append(handle)
        legend_labels.append(label)
    return ax, legend_handles, legend_labels


def _get_legend_handles_labels(legend: Legend) -> tuple[list[Any], list[str]]:
    handles = legend.legendHandles
    labels = legend.get_texts()
    labels = [label.get_text() for label in labels]
    return handles, labels


def _safe_get_legend_handles_labels(ax: Axes) -> tuple[list[Any], list[str]]:
    # First, get the handles and labels from the axis, if exists
    if isinstance(ax.legend_, Legend):
        handles, labels = _get_legend_handles_labels(ax.legend_)
        if len(handles) > 0:
            return handles, labels

    # If we didn't find handles, try to get them from the figure/axes
    handles, labels = ax.get_legend_handles_labels()
    return handles, labels


def _safe_draw_legend(
    ax: Axes, handles: list[Any], labels: list[str], **legend_kwargs: Any
) -> tuple[Legend, Axes]:
    legend_kwargs = legend_defaults | legend_kwargs
    ax, legend_handles, legend_labels = update_legend_handles_labels(
        ax, handles, labels
    )
    leg = ax.legend(legend_handles, legend_labels, **legend_kwargs)
    ax.legend_ = leg
    return leg, ax


def plot_abs_mag(
    sn: SN,
    ax: Optional[Axes] = None,
    split_by_site: bool = False,
    label_sites: bool = False,
    mark_unsub: bool = False,
    **plot_kwargs: Any,
) -> Axes:

    if ax is None:
        fig, ax = make_single_figure()

    if split_by_site:
        ax = plot_split_by_site(
            sn,
            ax,
            plot_type=plot_mag,
            absmag=True,
            label_sites=label_sites,
            mark_unsub=mark_unsub,
            **plot_kwargs,
        )
        return ax

    ax = plot_all_bands(
        sn,
        ax,
        site="all",
        plot_type=plot_mag,
        absmag=True,
        mark_unsub=mark_unsub,
        **plot_kwargs,
    )
    return ax


def label_axis(
    ax: Axes,
    xlabel: str,
    ylabel: str,
    legend: bool = True,
    custom_labels: Optional[list[str]] = None,
    custom_handles: Optional[list[Line2D | Patch | Any]] = None,
    **legend_kwargs: Any,
) -> tuple[Axes, Optional[Legend]]:
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)  # fontweigt = 'semibold'
    if legend:  # draw legend
        legend_kwargs = legend_defaults | legend_kwargs

        # Safely get handles and labels from the axis
        # Append custom handles and labels if given
        # Then (re)draw the legend
        if custom_labels is not None and custom_handles is not None:
            leg, ax = _safe_draw_legend(
                ax, custom_handles, custom_labels, **legend_kwargs
            )
            return ax, leg

        handles, labels = _safe_get_legend_handles_labels(ax)
        leg, ax = _safe_draw_legend(ax, handles, labels, **legend_kwargs)
        return ax, leg
    return ax, None


def format_axis(
    ax: Axes,
    invert: bool = False,
    log: bool = False,
    xlim: Optional[tuple[float, float]] = None,
    ylim: Optional[tuple[float, float]] = None,
    x_tick_interval: Optional[float] = None,
    y_tick_interval: Optional[float] = None,
    x_minors: Optional[int] = 5,
    y_minors: Optional[int] = 5,
) -> Axes:
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)
    if x_tick_interval is not None:
        ax.xaxis.set_major_locator(MultipleLocator(x_tick_interval))
    if y_tick_interval is not None:
        ax.yaxis.set_major_locator(MultipleLocator(y_tick_interval))
    if x_minors is not None:
        ax.xaxis.set_minor_locator(AutoMinorLocator(x_minors))
    if x_minors is not None:
        ax.yaxis.set_minor_locator(AutoMinorLocator(y_minors))
    if invert:
        ax.invert_yaxis()
    if log:
        ax.set_yscale("log")
    return ax


def plot_emcee(res: MinimizerResult):
    emcee_plot = corner.corner(
        res.flatchain,
        labels=res.var_names,
        truths=list(res.params.valuesdict().values()),
    )
    return emcee_plot

# class GenericPlot:
#     def __init__(self, f: PlotType):
#         self.f = f

#     def __call__(self, *args: Any, **kwargs: Any) -> Axes:
#         return self.f(*args, **kwargs)

class SNPlotTypes(Enum):
    ABS_MAG = plot_abs_mag
    LIM = plot_lims
    BAND = plot_band
    BANDS = plot_all_bands
    SITES = plot_split_by_site

    @classmethod
    def values(cls):
        return cls.__dict__.values()

    def __call__(self, *args: Any, **kwargs: Any) -> Axes:
        return self.value(*args, **kwargs)

class PhotPlotTypes(Enum):
    MAG = plot_mag
    LIM = plot_lim
    FLUX = plot_flux
    LUM = plot_lum

    @classmethod
    def values(cls):
        return cls.__dict__.values()


@dataclass
class PlotParameters:
    absmag: bool = False
    site: str = "all"  # or Site | None
    band: Filter = Filter("Unknown")
    shift: float = 0
    error_snake: bool = False
    as_flux: bool = False
    make_labels: bool = True
    label_sites: bool = False
    split_by_site: bool = True
    colors: ColorIterable = field(default_factory=lambda: DEFAULT_COLORS)
    shifts: Optional[list[float]] = None
    mark_unsub: bool = False

    def get(self, key: str) -> Any:
        return getattr(self, key)

    def items(self) -> Iterable[tuple[str, Any]]:
        d = asdict(self)
        d["band"] = self.band
        return d.items()


@dataclass
class AxisParameters:
    xlabel: str = "Phase (days)"
    ylabel: str = "Magnitude"
    legend: bool = True
    legend_kwargs: Optional[dict[str, Any]] = None
    invert: bool = False
    log: bool = False
    xlim: Optional[tuple[float, float]] = None
    ylim: Optional[tuple[float, float]] = None
    x_tick_interval: Optional[float] = None
    y_tick_interval: Optional[float] = None
    x_minors: Optional[int] = 5
    y_minors: Optional[int] = 5
    custom_labels: Optional[list[str]] = None
    custom_handles: Optional[list[Line2D | Patch | Any]] = None

    def get(self, key: str) -> Any:
        return getattr(self, key)

    def items(self) -> Iterable[tuple[str, Any]]:
        return asdict(self).items()

    def format_axis(self, ax: Axes):
        kwargs = self.legend_kwargs or legend_defaults
        ax, leg = label_axis(
            ax,
            self.xlabel,
            self.ylabel,
            self.legend,
            custom_handles=self.custom_handles,
            custom_labels=self.custom_labels,
            **kwargs,
        )
        ax = format_axis(
            ax,
            self.invert,
            self.log,
            self.xlim,
            self.ylim,
            self.x_tick_interval,
            self.y_tick_interval,
            self.x_minors,
            self.y_minors,
        )
        return ax


class Plotter:
    def __init__(
        self,
        sn: SN,
        plot_type: PlotType = SNPlotTypes.ABS_MAG,
        params: PlotParameters = PlotParameters(),
        axis_params: AxisParameters = AxisParameters(),
        **plot_kwargs: Any,
    ):
        self.sn = sn
        self.sn.sites.markers = cast(
            dict[int, str | None], get_site_markers(sn.sites.markers)
        )
        self.plot_type = plot_type
        self.signature = list(inspect.signature(self.plot_type).parameters.keys())
        self.plot_params = params
        self.plot_kwargs = plot_kwargs
        self.axis_params = axis_params

    def update_plot_type(self, plot_type: PlotType, **plot_kwargs: Any):
        self.plot_type = plot_type
        self.signature = list(inspect.signature(plot_type).parameters.keys())
        self.plot_kwargs = plot_kwargs

    def get_sn_or_phot(self) -> SN | Photometry:
        if self.plot_type in SNPlotTypes.values():
            return self.sn
        elif self.plot_type in PhotPlotTypes.values():
            lims = self.plot_type == PhotPlotTypes.LIM
            return self.sn.band(
                self.plot_params.band.name,
                site=self.plot_params.site,
                return_absmag=self.plot_params.absmag,
                flux=self.plot_params.as_flux,
                lims=lims,
            )
        raise ValueError(f"Plot type {self.plot_type} not recognised")

    def get_plot_params(self) -> dict[str, Any]:
        kwargs = {k: v for k, v in self.plot_params.items() if k in self.signature}
        return kwargs | self.plot_kwargs

    def save_current_legend(self, ax: Axes):
        if ax.get_legend() is not None:
            h, l = ax.get_legend_handles_labels()
            if self.axis_params.custom_labels is None:
                self.axis_params.custom_labels = l
            if self.axis_params.custom_handles is None:
                self.axis_params.custom_handles = h
            ax, h, l = update_legend_handles_labels(
                ax, self.axis_params.custom_handles, self.axis_params.custom_labels
            )
            self.axis_params.custom_handles = h
            self.axis_params.custom_labels = l

    def plot(
        self, ax: Optional[Axes] = None, **fig_kwargs: Any
    ) -> tuple[Figure | None, Axes]:
        kwargs = self.get_plot_params()
        arg = self.get_sn_or_phot()
        fig = None
        if ax is None:
            fig, ax = make_single_figure(**fig_kwargs)
        ax = self.plot_type(arg, ax=ax, **kwargs)
        self.save_current_legend(ax)
        return fig, ax

    def label_axis(
        self,
        ax: Axes,
        xlabel: str,
        ylabel: str,
        legend: bool = True,
        **legend_kwargs: Any,
    ) -> tuple[Axes, Optional[Legend]]:
        self.axis_params = replace(
            self.axis_params,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            legend_kwargs=legend_kwargs,
        )
        return label_axis(ax, xlabel, ylabel, legend, **legend_kwargs)

    def format_axis(
        self,
        ax: Axes,
        invert: bool = False,
        log: bool = False,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        x_tick_interval: Optional[float] = None,
        y_tick_interval: Optional[float] = None,
        x_minors: Optional[int] = 5,
        y_minors: Optional[int] = 5,
    ) -> Axes:
        self.axis_params = replace(
            self.axis_params,
            invert=invert,
            log=log,
            xlim=xlim,
            ylim=ylim,
            x_tick_interval=x_tick_interval,
            y_tick_interval=y_tick_interval,
            x_minors=x_minors,
            y_minors=y_minors,
        )
        return format_axis(
            ax,
            invert,
            log,
            xlim,
            ylim,
            x_tick_interval,
            y_tick_interval,
            x_minors,
            y_minors,
        )

    def plot_emcee(self, res: MinimizerResult):
        return plot_emcee(res)

    def setup_plot(self):
        sn = self.sn
        # Can also set automatic plot shifts & colors, but here using custom ones.
        sn = make_plot_shifts(
            sn=sn,
            reversed_for_abs=self.plot_params.absmag,
            shifts=self.plot_params.shifts,
        )
        sn = make_plot_colors(sn=sn, colors=self.plot_params.colors)
        self.sn = sn

    def legend_additions(self, ax: Axes) -> None:
        if self.plot_params.mark_unsub:
            handles, labels = get_unsub_mark()
            ax, handles, labels = update_legend_handles_labels(ax, handles, labels)
            if self.axis_params.custom_labels is None:
                self.axis_params.custom_handles = handles
                self.axis_params.custom_labels = labels
            else:
                for handle, label in zip(handles, labels):
                    if label not in self.axis_params.custom_labels:
                        self.axis_params.custom_labels.append(label)
                        if self.axis_params.custom_handles is None:
                            raise ValueError(
                                f"Custom handles should not be None when custom labels are {self.axis_params.custom_labels}"
                            )
                        self.axis_params.custom_handles.append(handle)

    def __call__(
        self, ax: Optional[Axes] = None, **fig_kwargs: Any
    ) -> tuple[Optional[Figure], Axes]:
        self.setup_plot()
        fig, ax = self.plot(ax, **fig_kwargs)

        ax = self.axis_params.format_axis(ax)
        return fig, ax



from pathlib import Path
import pandas as pd
from astrosn.supernova import SN, SNSerializer
from astrosn.photometry import PhotFactory, Photometry
from typing import Sequence, Optional, Type, Callable
from .readers import PathType, resolve_path, PhotReader
from .converters import Converter
NEEDED_KEYS = 'site,band,mag,mag_err,jd,sub'.split(',')
Collator = Callable[[PathType], Photometry]


def collate_phot(path: PathType, converter: Type[Converter], reader: PhotReader) -> Photometry:
    path = resolve_path(path)
    df = pd.concat([reader(p) for p in path.glob(converter.glob_str)], ignore_index=True)
    phot_df = converter(df=df).convert()
    return PhotFactory.from_df(phot_df)


add_phot = PhotFactory.add_phot


def create_sn(name: str, dir_path: PathType, collator: Collator,
              redshift: float = 0, phase_zero: float = 0,
              save_sn: bool = True) -> SN:
    """
    Create a Supernova object from a directory of photometry files,
    optionally saving it to the parent directory.
    """
    phot = collator(dir_path)
    sn = SN.from_phot(name=name, phot=phot, redshift=redshift, phase_zero=phase_zero)
    if save_sn:
        SNSerializer(sn).to_csv(Path(dir_path).parent)
    return sn


def update_sn(sn: SN, phot: Photometry) -> SN:
    # Add new sites.
    new_sites = set(phot.site.unique()) - set(sn.sites.site_ids)
    if new_sites != set():
        for site in new_sites:
            sn.add_site(name=str(site), site_id=site)

    # Add new bands.
    sn.bands = sn.make_bands(list(sn.phot.band.unique()), ebv=sn.sninfo.ebv)
    sn.set_phases()  # also recalculates the old phases.
    return sn


def verify_columns(df: pd.DataFrame, keys: Optional[Sequence['str']] = None) -> bool:
    if keys is None:
        keys = NEEDED_KEYS
    return set(keys) - set(df.columns) == set()



from typing import Any, MutableMapping, Optional, Callable, Sequence, cast
from astropy.table import Table
from astropy.time import Time
import pandas as pd
from pathlib import Path
import warnings

PathType = Path | str
PhotReader = Callable[[PathType], pd.DataFrame]


def resolve_path(path: PathType) -> Path:
    return Path(path).resolve(strict=True)


def read_astropy_table(path: PathType, fmt: str = 'ascii.ecsv') -> Table:
    path = resolve_path(path)
    return Table.read(path, format=fmt)


def read_pandas_csv(path: PathType, index_col: Optional[int | str] = None,
                    is_series: bool = False) -> pd.DataFrame:
    path = resolve_path(path)
    df = pd.read_csv(path, index_col=index_col)
    return df if not is_series else df.squeeze('columns')


def read_astropy_table_as_df_with_times_as_jd(path: PathType, fmt: str = 'ascii.ecsv') -> pd.DataFrame:
    at = read_astropy_table(path, fmt=fmt)
    at = times_as_jd(at)
    return at.to_pandas()


def get_astropy_times_as_jd(at: Table) -> Table:
    times_cols = get_times_cols(at)

    if len(times_cols) == 0:
        warnings.warn(f"No time columns found in astropy table with columns: {at.colnames}", UserWarning)
        return at

    for col in times_cols:
        fmt = detect_mjd_or_jd(cast(Sequence, at[col]))
        if fmt is None:
            warnings.warn(f"Could not determine time format for column: {col}.", UserWarning)
        else:
            at[col] = Time(at[col], format=fmt).jd
    return at

def detect_mjd_or_jd(col: Sequence[Any]) -> Optional[str]:
    def mjd(sval: str) -> bool:
        return len(sval) == 5 and sval[0] in ['5', '6']

    def jd(sval: str) -> bool:
        return len(sval) == 7 and sval.startswith("245")

    mjds = []
    jds = []
    for val in col:
        sval = str(int(val))
        if mjd(sval):
            mjds.append(val)
        elif jd(sval):
            jds.append(val)
        if len(mjds) > 1 and len(jds) > 1:
            break
    if len(mjds) == len(col):
        return 'mjd'
    elif len(jds) == len(col):
        return 'jd'
    return None


def get_times_cols(at: Table) -> list[str]:
    time_names = ["time", "jd", "mjd"]
    return [col for col in at.colnames if col.strip().lower() in time_names]


def times_as_jd(at: Table) -> Table:
    for name, col in at.columns.items():
        if isinstance(col, Time):
            at[name] = col.jd
    return at



from __future__ import annotations
from dataclasses import dataclass, asdict, field, fields
from typing import (
    Any,
    Generic,
    Sequence,
    TypeVar,
    Protocol,
    Mapping,
    Iterable,
    Hashable,
    overload,
    runtime_checkable,
)
import warnings

import numpy as np
import pandas as pd
from pandas import Series, DataFrame
#from pandera.typing import Series, DataFrame
from astropy.units import Quantity
import astropy.units as u

from .filters import Filter
from .utils import Number, get_field_dtype_default, _is_quantity
T = TypeVar("T")
GenericType = TypeVar("GenericType", covariant=True)
NumType = TypeVar('NumType', float, int, Quantity)
class Series(Series, Generic[GenericType]):
    ...


ArrayLike = TypeVar("ArrayLike", np.ndarray[Any, Any], Sequence[Number], Series)
N = TypeVar("N", Quantity, float, int, Number)

SeriesSite = Series[str] | Series[int]
DFDict = Mapping[str, ArrayLike] | Mapping[Hashable, Any]


@dataclass
class HasPhot(Protocol):
    jd: Series[float]
    band: Series[str]
    sub: Series[bool]
    site: SeriesSite
    phase: Series[float]


@dataclass
@runtime_checkable
class HasMag(Protocol):
    mag: Series[float]
    mag_err: Series[float]

    def absmag(self, dm: N, ext: N) -> Series[float]:
        ...


@dataclass
@runtime_checkable
class HasFlux(Protocol):
    flux: Series[float]
    flux_err: Series[float]


@dataclass
@runtime_checkable
class _HasFlux(HasFlux, Protocol):
    @staticmethod
    def mag_to_flux(mag: Series[float], filt: Filter) -> Series[float]:
        ...

    @staticmethod
    def mag_to_flux_err(mag_err: Series[float], flux: Series[float]) -> Series[float]:
        ...


@dataclass
@runtime_checkable
class HasMagFlux(HasMag, _HasFlux, Protocol):
    @classmethod
    def from_magphot(cls, magphot: HasMag, band: Filter) -> "HasMagFlux":
        ...


@dataclass
@runtime_checkable
class HasLuminosity(Protocol):
    lum: Series[float]
    lum_err: Series[float]


@dataclass
@runtime_checkable
class HasLumFlux(HasLuminosity, HasFlux, Protocol):
    ...


@dataclass
@runtime_checkable
class HasBlackBody(HasLuminosity, Protocol):
    radius: Series[float]
    radius_err: Series[float]
    temp: Series[float]
    temp_err: Series[float]


@dataclass
@runtime_checkable
class HasBlackBodyFlux(HasBlackBody, HasFlux, Protocol):
    ...


@dataclass
class PhotMethods(Protocol):
    def calc_phases(self, phase_zero: N) -> None:
        ...

    def restframe_phases(self, redshift: N) -> Series[float]:
        ...

    def as_dataframe(self) -> DataFrame:
        ...

    def __len__(self) -> int:
        ...

    @classmethod
    def from_dict(cls, d: DFDict[Any]) -> "Photometry":
        ...

    def masked(self, cond: Iterable[bool]) -> "Photometry":
        ...

    def sorted(self) -> "Photometry":
        ...

@runtime_checkable
class ImplementsMag(HasPhot, HasMag, PhotMethods, Protocol):
    pass

@runtime_checkable
class ImplementsFlux(HasPhot, _HasFlux, PhotMethods, Protocol):
    pass

@runtime_checkable
class ImplementsMagFlux(HasPhot, HasMagFlux, PhotMethods, Protocol):
    pass

@runtime_checkable
class ImplementsLuminosity(HasPhot, HasLuminosity, PhotMethods, Protocol):
    pass

@runtime_checkable
class ImplementsBlackBody(HasPhot, HasBlackBody, PhotMethods, Protocol):
    pass

Photometry = (
    ImplementsMag
    | ImplementsFlux
    | ImplementsMagFlux
    | ImplementsLuminosity
    | ImplementsBlackBody
)

FluxPhotometry = ImplementsFlux | ImplementsMagFlux 
MagPhotometry = ImplementsMag | ImplementsMagFlux


P = TypeVar('P', bound=Photometry)


@dataclass
class BasePhotMixin(PhotMethods, Generic[P]):
    jd: Series[float] = Series(dtype=float)
    band: Series[str] = Series(dtype=str)
    phase: Series[float] = Series(dtype=float)
    sub: Series[bool] = Series(dtype=bool)
    site: Series[int] | Series[str] = Series(dtype=int)
    # only for backwards compatibility do not access directly as it overrides builtin filter.
    filter: Series[str] = Series(dtype=str)

    def __post_init__(self) -> None:
        if 0 < len(self.filter) == len(self):
            warnings.warn(
                "Phot: `filter` is deprecated, use `band` instead", DeprecationWarning
            )
            self.band = self.filter

        if len(self.band) != len(self):
            raise ValueError(
                "Either band or filter must be given the same length as jd."
                f"Got band: {len(self.band)}, "
                f"filter: {len(self.filter)}, jd: {len(self)}"
            )
        self.filter = self.band
        # @TODO: remove this try/except after tests.
        try:
            self._fill_missing()
        except Exception as e:
            warnings.warn(f"Could not fill missing values of {self.__class__}: {e}")

    def _fill_missing(self) -> None:
        if len(self) == 0:
            return
        for _field in fields(self):

            if len(getattr(self, _field.name)) == 0:
                dtype, default = get_field_dtype_default(_field)
                setattr(
                    self,
                    _field.name,
                    Series([default] * len(self), dtype=dtype, name=_field.name),
                )

    @classmethod
    def from_dict(cls, d: DFDict[Any]) -> P:
        if cls == BasePhotMixin:
            raise TypeError("Cannot instantiate Base class. Use a subclass.")
        return cls(**{k: v for k, v in d.items() if k in [f.name for f in fields(cls)]})  # type: ignore

    def masked(self, cond: Iterable[bool]) -> P:
        d2 = {name: value[cond] for name, value in asdict(self).items()}
        return self.from_dict(d2)

    def calc_phases(self, phase_zero: float | int) -> None:
        self.phase = self.jd - phase_zero

    def restframe_phases(self, redshift: float | int) -> Series[float]:
        if self.phase is None:
            raise AttributeError(
                "self.phase must not be None, "
                "calculate it first using calc_phases with a phase_zero"
            )
        return self.phase / (1 + redshift)

    def as_dataframe(self) -> DataFrame:
        return DataFrame(asdict(self))

    def __len__(self) -> int:
        return len(self.jd)

    def sorted(self) -> "Photometry":
        df = self.as_dataframe().sort_values(by="jd")
        df.reset_index(drop=True, inplace=True)
        return self.from_dict(df.to_dict("series"))



class _MagMethodsMixin:
    mag: Series[float]

    def absmag(self, dm: N, ext: N) -> Series[N]:
        return self.mag - dm - ext

@dataclass
class MagMixin:
    mag: Series[float] = Series(dtype=float)
    mag_err: Series[float] = Series(dtype=float)

@dataclass
class MagPhot(BasePhotMixin[ImplementsMag], MagMixin, _MagMethodsMixin, ImplementsMag):
    pass


class _FluxMethodsMixin:
    
    @staticmethod
    def mag_to_flux(mag: Series[float], filt: Filter) -> Series[float]:
        return 10 ** (mag / -2.5) * filt.zp

    @staticmethod
    def mag_to_flux_err(mag_err: Series[float], flux: Series[float]) -> Series[float]:
        return 2.303 * mag_err * flux

@dataclass
class FluxMixin:
    flux: Series[float] = Series(dtype=float)
    flux_err: Series[float] = Series(dtype=float)

@dataclass
class FluxPhot(BasePhotMixin[ImplementsFlux], FluxMixin, _FluxMethodsMixin, ImplementsFlux):
    pass


@dataclass
class Phot(BasePhotMixin[ImplementsMagFlux], MagMixin, FluxMixin, _MagMethodsMixin, _FluxMethodsMixin, ImplementsMagFlux, ImplementsFlux, ImplementsMag):


    @classmethod
    def from_magphot(cls, magphot: HasMag, band: Filter) -> ImplementsMagFlux:
        flux = cls.mag_to_flux(magphot.mag, band)
        flux_err = cls.mag_to_flux_err(magphot.mag_err, flux)
        magdict = asdict(magphot)
        fluxdict = {"flux": flux, "flux_err": flux_err}
        return cls.from_dict(magdict | fluxdict)

@dataclass
class LumMixin:
    lum: Series[float] = Series(dtype=float)
    lum_err: Series[float] = Series(dtype=float)

@dataclass
class LumPhot(BasePhotMixin[ImplementsLuminosity], LumMixin, FluxMixin, ImplementsLuminosity, HasLumFlux):
    pass

@dataclass
class BBLumPhot(BasePhotMixin[ImplementsBlackBody], LumMixin, FluxMixin, ImplementsBlackBody, HasBlackBodyFlux):
    radius: Series[float] = Series(dtype=float)
    radius_err: Series[float] = Series(dtype=float)
    temp: Series[float] = Series(dtype=float)
    temp_err: Series[float] = Series(dtype=float)


class PhotFactory:
    """
    Factory class to create Photometry objects.
    """

    def __init__(self, sn_phot: Photometry) -> None:
        self.sn_phot = sn_phot

    @staticmethod
    def from_df(df: DataFrame) -> Photometry:
        #d = cast(Mapping[str, Series], df.to_dict("series"))
        d = df.to_dict("series")
        mag = "mag" in d
        flux = "flux" in d
        lum = "lum" in d
        bb = "radius" in d and "temp" in d
        if bb:
            return BBLumPhot.from_dict(d)
        if lum:
            return LumPhot.from_dict(d)
        if mag and flux:
            return Phot.from_dict(d)
        elif mag:
            return MagPhot.from_dict(d)
        elif flux:
            return FluxPhot.from_dict(d)
        raise ValueError("df must contain either 'mag' or 'flux' columns")

    @classmethod
    def add_phot(cls, sn_phot: Photometry, new_phot: Photometry) -> Photometry:
        """
        Add photometry to a Supernova.
        """
        df = pd.concat(
            [sn_phot.as_dataframe(), new_phot.as_dataframe()], ignore_index=True
        )
        df = df[sn_phot.as_dataframe().columns.drop("filter")]
        return cls.from_df(df)

    def concat_phot(self, new_phot: Photometry) -> Photometry:
        """
        Add photometry to a Supernova.
        """
        self.sn_phot = self.add_phot(self.sn_phot, new_phot)
        return self.sn_phot

    def extend_phot(self, new_phot_type: type[Photometry]) -> Photometry:
        """
        Extend the photometry of a Supernova to a new type.
        """
        d = self.sn_phot.as_dataframe().to_dict("series")
        self.sn_phot = new_phot_type.from_dict(d)
        return self.sn_phot



from pathlib import Path
from .flows import collate_flows
from .readers import PathType
from .ztf import collate_ztf_flows, collate_ztfphot, collate_ztfphot_limits
from .utils import add_phot
from .collators import AbstractCollator, collate_ecsv, collate_csv, collate_json
from astrosn.supernova import SN, SNInfo
from astrosn.photometry import Photometry
from astrosn.sites import Sites, flows_sites, SiteDict
from enum import Enum
from typing import Generator, Mapping, Optional, Any
from tempfile import TemporaryDirectory
from contextlib import contextmanager


PREDEFINED_COLLATORS = {
    "flows": collate_flows,
    "ztf_flows": collate_ztf_flows,
    "ztfphot": collate_ztfphot,
}
PREDEFINED_LIMITS_COLLATORS = {"ztfphot_limits": collate_ztfphot_limits}
GENERIC_COLLATORS = {"ecsv": collate_ecsv, "csv": collate_csv, "json": collate_json}


class DataIngestor:
    """
    Base class for data ingestors.
    """

    def __init__(
        self,
        path: PathType,
        collators: Optional[Mapping[str, AbstractCollator]] = None,
        limits_collators: Optional[Mapping[str, AbstractCollator]] = None,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = None,
        **sninfo_kwargs: Any,
    ) -> None:
        """
        A class to collect data from a directory of files and create an SN object.
        RA and Dec can be provided in degrees, or they can be looked up from the SN name,
        which should match IAU specifications.Redshift can be given, if known.
        """
        self.path = Path(path).resolve()
        self.tmp_path = self.path
        self.files = self.get_files()
        self.processed_files = []
        if len(self.files) == 0:
            raise FileNotFoundError(f"No files found in {self.path}.")

        if sninfo is None:
            sninfo = SNInfo.from_csv(self.path)
        self.sninfo = sninfo
        self.sitemap = sitemap
        self.collators = collators or PREDEFINED_COLLATORS
        self.limits_collators = limits_collators or PREDEFINED_LIMITS_COLLATORS
        self.sninfo_kwargs = sninfo_kwargs

    def setup(self) -> None:

        """
        Setup the data ingestor
        """
        pass

    def get_files(self) -> list[Path]:
        return list(self.path.glob("*"))

    @contextmanager
    def tmpdir(self) -> Generator[TemporaryDirectory, None, None]:
        _tmpdir = TemporaryDirectory()
        self.tmp_path = Path(_tmpdir.name)
        try:
            yield _tmpdir
        finally:
            _tmpdir.cleanup()

    def move_files_to_temp_dir(self, tmpdir_path: Path) -> None:
        for file in self.files:
            temp_file = tmpdir_path / file.name
            temp_file.symlink_to(file)

    def process_file(self, file: Path, unlink: bool = True) -> None:
        if file in self.processed_files:
            return
        if file.is_symlink() and unlink:
            file.unlink()
        self.processed_files.append(file)

    def load_phot(self, lims: bool = False, unlink: bool = True) -> list[Photometry]:
        """
        Load the photometry from the data in the directory.
        Unlink the files after they are processed.
        """
        phots = []
        phot_type = "limits" if lims else "photometry"
        collators = self.limits_collators if lims else self.collators
        for source, collator in collators.items():
            try:
                phots.append(collator(self.tmp_path))
                print(f"Loaded {phot_type} from {source}.")

                for file in self.tmp_path.glob(collator.converter.glob_str):
                    self.process_file(file, unlink=unlink)

            except FileNotFoundError:
                print(f"Did not find {phot_type} from {source}.")
            if len(phots) == 2:
                phots.append(add_phot(phots.pop(0), phots.pop(0)))

        return phots

    def load_sn(self) -> SN:
        """
        Load the SN object from the data in the directory.
        But first symlink all the files to a temporary directory,
        so that the collators can delete the files they process.
        """
        with self.tmpdir() as tmpdir:
            self.move_files_to_temp_dir(Path(tmpdir.name))
            phots = self.load_phot()
            if self.limits_collators is not None:
                lims = self.load_phot(lims=True)
            else:
                lims = []
            if len(phots) == 0:
                raise FileNotFoundError(
                    f"No photometry found in {self.path}."
                    f"With defined data collators: {self.collators}"
                )

            phot = phots[0]
            sites = None
            if self.sitemap:
                sites = Sites.from_sitemap(
                    {s: self.sitemap.get(s, str(s)) for s in phot.site.unique()}
                )

            phot = phot.sorted()
            return SN.from_phot(
                phot=phot,
                name=self.sninfo.name,
                redshift=self.sninfo.redshift,
                sub_only=self.sninfo.sub_only,
                phase_zero=self.sninfo.phase_zero,
                limits=lims[0] if len(lims) > 0 else None,
                sninfo=self.sninfo,
                sites=sites,
                **self.sninfo_kwargs,
            )

    @staticmethod
    def query_user(query: str) -> str:
        return input(query)


class GenericIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = None,
        **sninfo_kwargs: Any,
    ) -> None:
        collators = GENERIC_COLLATORS
        super().__init__(path, collators, None, sninfo, sitemap, **sninfo_kwargs)


class ECSVIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = None,
        **sninfo_kwargs: Any,
    ) -> None:
        collators = {"ecsv": collate_ecsv}
        super().__init__(path, collators, None, sninfo, sitemap, **sninfo_kwargs)


class CSVIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = None,
        **sninfo_kwargs: Any,
    ) -> None:
        collators = {"csv": collate_csv}
        super().__init__(path, collators, None, sninfo, sitemap, **sninfo_kwargs)


class JSONIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = None,
        **sninfo_kwargs: Any,
    ) -> None:
        collators = {"json": collate_json}
        super().__init__(path, collators, None, sninfo, sitemap, **sninfo_kwargs)


class FlowsIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = flows_sites,
        **sninfo_kwargs: Any,
    ) -> None:
        collators = PREDEFINED_COLLATORS
        limits_collators = PREDEFINED_LIMITS_COLLATORS

        super().__init__(
            path, collators, limits_collators, sninfo, sitemap, **sninfo_kwargs
        )


class ZTFIngest(DataIngestor):
    def __init__(
        self,
        path: PathType,
        sninfo: Optional[SNInfo] = None,
        sitemap: Optional[SiteDict] = {0 : "ZTF"},
        **sninfo_kwargs: Any,
    ) -> None:
        collators = {"ztf": collate_ztf_flows, 'ztfphot': collate_ztfphot}
        limits_collators = None
        super().__init__(
            path, collators, limits_collators, sninfo, sitemap, **sninfo_kwargs
        )


class Ingestors(Enum):
    flows = FlowsIngest
    generic = GenericIngest
    ecsv = ECSVIngest
    csv = CSVIngest
    json = JSONIngest
    ztf = ZTFIngest


def ingest_sn(
    path: PathType,
    snname: str = "SN1990Unknown",
    ingestor: str | Ingestors = Ingestors.flows,
    sninfo: Optional[SNInfo] = None,
    sitemap: Optional[SiteDict] = None,
    **sninfo_kwargs: Any,
) -> SN:
    """
    Ingest a SN from a directory of data.

    Parameters
    ----------
    path: str | Path
        The path to the directory of data.
    snname: str
        The name of the SN.
    ingestor: str | Ingestors = Ingestors.flows
        The ingestor to use. Can be a string or an Ingestors enum. Flows is the default.
    sninfo: Optional[SNInfo] = None
        An SNInfo object to use for the SN. If None, will create via query.
    sitemap: Optional[dict[int, str]] = None
        A dictionary mapping site numbers to site names. Uses FLOWS sites by default if ingestor is flows.
        Pass None to have them automatically generated from the files. (Will check for file ending in _sites,
        and if not found, will use the site numbers in the photometry file. If photometry file has sitenames
        instead of numbers, will use those.)
        Else pass a dictionary mapping site numbers to site names. Remember that photometry files must have
        the corresponding site numbers for each telescope in every row.
    sninfo_kwargs: Any
        Any additional kwargs to pass to the SNInfo constructor.
        Such as redshift, phase_zero, sub_only, etc. Redshift will be looked up from flows if possible and not passed.
        tendrils must be installed for this.
    """
    if isinstance(ingestor, str):
        ingestor = Ingestors[ingestor]

    redshift = sninfo_kwargs.pop("redshift", 0.0)
    if isinstance(redshift, str):
        redshift = float(redshift)  # try to convert to float
    if not isinstance(redshift, float):
        raise TypeError(
            f"redshift must be a float or a string that can be converted to a float. Got {redshift}."
        )
    #if redshift == 0.0:
        

    if sninfo is None:
        sninfo = SNInfo.from_name(snname, redshift=redshift, **sninfo_kwargs)

    if ingestor == Ingestors.flows:
        sitemap = flows_sites if sitemap is None else sitemap

    return ingestor.value(path, sninfo, sitemap, **sninfo_kwargs).load_sn()



from numpy.typing import NDArray
import numpy as np
from astrosn import SN, Photometry, Filter
from astrosn.utils import PathType, StrEnum
from typing import Any, Protocol, TypeVar

# Filters
bolo_filt = Filter('Bolo')
blackbody_filt = Filter("BB")

class BoloType(StrEnum):
    """Bolometric luminosity type."""
    quasi = 'Bolo'  # Quasi-bolometric
    blackbody = 'BB'  # Blackbody


def bolo_weighted_xyz(sn: SN,
                      band: BoloType = BoloType.quasi) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Get the bolometric luminosity weighted x, y, z coordinates."""
    sn.set_phases()
    snr = sn.restframe()
    phot = snr.band(band)
    weights = 1./phot.lum_err.values
    return phot.phase.values, phot.lum.values, weights


class ModelFit(Protocol):

    def __init__(self, params: Any, **kwargs):
        ...

    def get_model(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        ...

    def to_json(self, path: PathType) -> None:
        ...



from dataclasses import asdict, dataclass, field
import json
from typing import Optional
import multiprocessing
import mpmath as mp
import numpy as np
from lmfit import minimize, Parameters, report_fit
from lmfit.minimizer import MinimizerResult
from numpy.typing import ArrayLike, NDArray

from astrosn.utils import PathType, StrEnum, FloatArr
from .bololc import BoloType, ModelFit, bolo_weighted_xyz


# Conversion Factors
days = 86400  # seconds
Msun = 1.98855e33  # grams
kms10 = 100000000.0  # 10000km/s -> cm/s

# Constants
Eni = 3.90e10  # ergs s-1 g-1
Eco = 6.78e9  # ergs s-1 g-1
tau_ni = 8.77  # *days #days * days =seconds
tau_co = 111.3  # *days #days * days =seconds

c = 2.998e10  # cm/s
Beta = 13.8  # constant
Kappa = 0.07  # cm2 g-1


# Derived quantities
# Diffusion Time
def get_tau_m(M_ej, v_ph):
    return (Kappa / (Beta * c)) ** (1 / 2) * (M_ej / v_ph) ** (1 / 2) / days  # taum in days


def get_s(tau_m):
    return tau_m * ((tau_co - tau_ni) / (2 * tau_co * tau_ni))  # days/days


def get_y(tau_m):
    return tau_m / (2 * tau_ni)  # unitless


def func_A(z, tau_m):
    y = get_y(tau_m)
    # return 2*z*np.exp((-2*z*y) + z**2)
    # print(2*z)
    TempA = mp.log(2 * z) - 2 * z * y + z ** 2
    # print(TempA)
    return mp.exp(TempA)


def func_B(z, tau_m):
    y = get_y(tau_m)
    s = get_s(tau_m)

    TempB = mp.log(2 * z) + 2 * z * s - 2 * z * y + z ** 2
    # print(TempB)
    return mp.exp(TempB)
    # return 2*z*np.exp(-2*z*y + 2*z*s + z**2)


def nonvector_L(t, *params):
    M_ej, M_ni, v_ph = params

    tau_m = get_tau_m(M_ej, v_ph)

    x = t / tau_m
    # Term1,er1=mp.quad(func_A,[0.0001,x],args=(tau_m))
    # Term2,er2=mp.quad(func_B,[0.0001,x],args=(tau_m))

    Term1 = mp.quad(lambda x: func_A(x, tau_m), [0.0001, x])
    Term2 = mp.quad(lambda x: func_B(x, tau_m), [0.0001, x])

    # lambda x: func_A(x, tau_m)

    # print M_ej/Msun,M_ni/Msun,v_ph/kms,Term1,Term2,get_y(tau_m),get_s(tau_m),tau_m,x

    lum = (M_ni * np.array(mp.exp(-x ** 2))) * ((Eni - Eco) * np.array(Term1) + (Eco * np.array(Term2))) / 1e43

    return lum * trapped(t * days, M_ej, v_ph)


L = np.vectorize(nonvector_L)


# EK_model*(trapped(xdata))
def get_Ek(M_ej, v_ph):
    return (3 / 10) * M_ej * (v_ph ** 2)


def tau(t, M_ej, Ek):
    T0 = np.sqrt((0.05305165 * Kappa * (M_ej ** 2)) / Ek)
    return (t / T0) ** (-2)


def trapped(t, M_ej, v_ph):
    Ek = get_Ek(M_ej, v_ph)
    return 1.0 - (0.965 * mp.exp(-tau(t, M_ej, Ek)))


def get_x_full(x: NDArray, E_exp: float) -> np.ndarray[float]:
    if E_exp == 0.0:
        x_full = x
    else:
        x_full = np.arange(x[0] - np.abs(E_exp) - 1.0, x[-1] + np.abs(E_exp), step=1.0)

    return x_full


def get_model_at_x(x, x_full, M, N, V, E_exp) -> np.ndarray[float]:
    model_now = L(x_full, M, N, V)

    model_now = model_now.astype(np.float64) * 1e43
    model_int = np.interp(x, x_full + E_exp, model_now)
    return model_int


def get_Ek_err(M_ej, M_ej_err, v_ph, v_ph_err):
    return get_Ek(M_ej, v_ph) * np.sqrt(((2*v_ph_err) / v_ph) ** 2 + (M_ej_err / M_ej) ** 2)


class RegularizedFit:

    def __init__(self, params: Parameters, varied_params: list[str],
                 lamb: float = 0):
        self.lamb = lamb
        self.params = params
        self.varied_params = varied_params
        self.regularization = 0

    def set_regularization(self, params, iter, resid, *args, **kws) -> None:
        n_samples = len(resid)
        param = np.array([params[p].value for p in self.varied_params])
        self.regularization = self.lamb * np.sum(param ** 2) / n_samples

    def regularized_least_squares_cost(self, r: ArrayLike) -> float:
        return np.sum(r ** 2) / 2 / len(r) + self.regularization


def process_params(params: Parameters) -> tuple[float, float, float, float]:
    mej = params['M_ej'].value * Msun
    mni = params['M_ni'].value * Msun
    e_exp = params['E_exp'].value
    vph = params['V_ph'].value * kms10
    return mej, mni, e_exp, vph


def get_model(params: Parameters, x: ArrayLike) -> NDArray[np.float64]:
    mej, mni, e_exp, vph = process_params(params)

    if abs(e_exp - 0.0) < 1e-3:
        # make +/- E_exp model range
        x_full = get_x_full(x, e_exp)
        x_full_p = x_full[x_full > 0.0]

        # calculate model within the full range shifted by E_exp
        model_int = get_model_at_x(x, x_full_p, mej, mni, vph, e_exp)
    else:
        model_now = L(x, mej, mni, vph)
        model_int = model_now.astype(np.float64) * 1e43

    return model_int

# fix velocity fit
def fit_func(params: Parameters, x: NDArray[np.float64], y: NDArray[np.float64],
             weights: Optional[NDArray[np.float64]] = None) -> NDArray[np.float64]:
    model_int = get_model(params, x)

    resids = model_int - y  # minimize this
    if weights is not None:
        resids = resids * weights
    return resids

@dataclass
class ArnettParams:
    mej: float
    vph: float
    e_exp: float = 0.0
    vary_e_exp: bool = True
    delta: float = 0.05

    e_exp_min: float = -3.0
    e_exp_max: float = 3.0
    e_exp_step: float = 0.2

    mej_min: float = 0.0
    mej_max: float = 10.0
    mej_step: float = 0.1

    delta_min: float = 0.00001
    delta_max: float = 0.99
    delta_step: float = 0.01

    @property
    def params(self) -> Parameters:
        params = Parameters()
        params.add('M_ej', value=self.mej, min=self.mej_min, max=self.mej_max,
                   vary=True, brute_step=self.mej_step)
        params.add('delta', value=self.delta, max=self.delta_max, min=self.delta_min, vary=True,
                   brute_step=self.delta_step)
        params.add('M_ni', expr='M_ej*delta', vary=False)
        params.add('E_exp', value=self.e_exp, max=self.e_exp_min, min=self.e_exp_max, vary=self.vary_e_exp,
                   brute_step=self.e_exp_step)
        params.add('V_ph', value=self.vph, vary=False)
        return params


class LMFitMethod(StrEnum):
    nelder = 'nelder'  # Nelder-Mead
    leastsq = 'leastsq'  # Levenberg-Marquardt
    least_squares = 'least_squares'  #  Trust Region Reflective
    bfgs = 'bfgs'  # Broyden-Fletcher-Goldfarb-Shanno
    brute = 'brute'  # brute force
    lbfgsb = 'lbfgsb'  # L-BFGS-B
    trust_constr = 'trust_constr'  # Trust Region Constrained
    slsqp = 'slsqp'  # Sequential Least SQuares Programming
    emcee = 'emcee'  # Markov Chain Monte Carlo


def fit_arnett(xdata: ArrayLike, ydata: ArrayLike, weights: Optional[ArrayLike] = None,
               arnett_params: ArnettParams = ArnettParams(1.0, 15, 0.0), regularization: float = 0,
               fit_emcee: bool = True, method: LMFitMethod | str = LMFitMethod.nelder) -> tuple[MinimizerResult, Optional[MinimizerResult]]:
    method = LMFitMethod(method)
    params = arnett_params.params
    fitter = RegularizedFit(params, ['M_ej', 'delta'], lamb=regularization)
    fit_result = minimize(fit_func, params, args=(xdata, ydata),
                          kws={'weights': weights}, method=method, nan_policy='omit',
                          iter_cb=fitter.set_regularization,
                          reduce_fcn=fitter.regularized_least_squares_cost)
    emcee_result = None
    if fit_emcee:
        # Explore posterior using emcee
        # fit_result.params.add('__lnsigma', value=np.log(0.1), min=np.log(0.001), max=np.log(2))
        is_weighted = weights is not None
        emcee_result = minimize(fit_func, fit_result.params, args=(xdata, ydata),
                                kws={'weights': weights}, method='emcee',
                                nan_policy='omit', is_weighted=is_weighted, progress=False,
                                burn=300, steps=1000, thin=20,
                                workers=max(min(multiprocessing.cpu_count()//2 + 3, 10), 1))

    return fit_result, emcee_result


@dataclass
class  ArnettModelFit(ModelFit):
    name: str = "Arnett"
    params: Parameters = field(default_factory=Parameters)

    mej: float = field(default=-999)
    mej_err: float = field(default=-999)
    mni: float = field(default=-999)
    mni_err: float = field(default=-999)
    eexp: float = field(default=-999)
    eexp_err: float = field(default=-999)
    vph: float = field(default=-999)
    vph_err: float = field(default=-999)
    ek: float = field(default=-999)
    ek_err: float = field(default=-999)

    def __post_init__(self):
        if self.mej == -999:
            self.set_pars()
            self.set_errors()
        self.params = self.build_params()

    def with_error(self, parname: str):
        return f"${getattr(self, parname):.2f} \pm {getattr(self, f'{parname}_err'):.3f}$"

    def with_error_1(self, parname: str):
        return f"${getattr(self, parname):.1f} \pm {getattr(self, f'{parname}_err'):.2f}$"

    def __getitem__(self, __name: str) -> str:
        if __name in self.params:
            return f"${self.params[__name].value} \pm {self.params[__name].stderr}$"
        raise KeyError(f"Parameter {__name} not found")

    def set_pars(self) -> None:
        self.mej = self.params['M_ej'].value
        self.mni = self.params['M_ni'].value
        self.eexp = self.params['E_exp'].value
        self.vph = self.params['V_ph'].value

    def set_errors(self) -> None:
        self.mej_err = self.params["M_ej"].stderr
        self.mni_err = self.params["M_ni"].stderr
        self.eexp_err = self.params["E_exp"].stderr
        self.vph_err = self.params["V_ph"].stderr
        self.ek = get_Ek(self.mej, self.vph)
        self.ek_err = get_Ek_err(self.mej, self.mej_err, self.vph, self.vph_err)

    def build_params(self) -> Parameters:
        params = Parameters()
        params.add('M_ej', value=self.mej, vary=True)
        params.add('delta', value=self.mni/self.mej, vary=True)
        params.add('M_ni', expr='M_ej*delta', vary=False)
        params.add('E_exp', value=self.eexp, vary=False)
        params.add('V_ph', value=self.vph, vary=False)

        for par in params:
            if par == 'delta':
                continue
            params[par].stderr = getattr(self, f"{par.replace('_','').lower()}_err")

        return params

    def get_model(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return get_model(self.params, x)

    def to_json(self, path: PathType) -> None:
        with open(path, 'w') as f:
            json.dump(asdict(self), f, indent=4)

    @classmethod
    def from_json(cls, path: PathType) -> "ArnettModelFit":
        with open(path, 'r') as f:
            data = json.load(f)
        return cls(**data)


def get_arnett_params(sn, phase_min: float = 0, phase_max: float = 10000, vph: float = 10., mej_init: float = 1) -> tuple[FloatArr, FloatArr, FloatArr, ArnettParams]:
    """Get the Arnett parameters for a SN
    vph: float = velocity of the ejecta in $10^4$ km/s
    mej_init: float = initial guess for the ejecta mass in solar masses
    """
    x, y, w = bolo_weighted_xyz(sn, BoloType.quasi)
    apars = ArnettParams(mej_init, vph=vph, vary_e_exp=False)

    mask = (x > phase_min) & (x < phase_max)
    return x[mask], y[mask], w[mask], apars



from functools import partial
from typing import Any, Optional, Iterable
import numpy as np
from numpy.typing import NDArray, ArrayLike
from astropy import constants  # type: ignore
import astropy.units as u
from astropy.modeling import fitting, physical_models
import multiprocessing
from astropy.cosmology import WMAP5  # type: ignore
from astropy.table import QTable
import pandas as pd
import tqdm
import warnings

from astrosn import Photometry, Filter, SN, LumPhot, BBLumPhot, PhotFactory
from astrosn.filters import FilterSorter
from .bololc import bolo_filt, blackbody_filt

# Constants
a = 4 * constants.sigma_sb / constants.c  # type: ignore
a_cg = a.to(u.erg / u.cm**3 / u.K**4)  # type: ignore
flam = u.erg / u.cm**2 / u.s / u.angstrom  # type: ignore
slam = flam / u.sr
DEFAULT_KWARGS = {"temperature": 10000 * u.K}
flux_units = u.erg / u.s / u.cm**2  # type: ignore
MyArrayLike = ArrayLike | u.Quantity




def bb_err(bb_fit: physical_models.BlackBody) -> u.Quantity:
    return (
        np.sqrt(
            ((bb_fit.scale.std) / bb_fit.scale) ** 2
            + (4 * bb_fit.temperature.std / bb_fit.temperature.value) ** 2  # type: ignore
        )
        * bb_fit.bolometric_flux
    )


@u.quantity_input(y=flam)
def init_bb_fit(
    y: flam, init_kwargs: Optional[dict[str, Any]] = None, normalize_flux: bool = False
) -> tuple[u.Quantity, physical_models.BlackBody, Optional[u.Quantity], float]:
    y = y / np.pi / u.sr
    norm = None
    if normalize_flux:
        norm = np.max(y)
        y = y / norm

    if init_kwargs is None:
        init_kwargs = DEFAULT_KWARGS

    scale_guess = np.mean(y)

    if normalize_flux:
        scale_guess = scale_guess << slam

    rescale = 1.0 / scale_guess.value
    init_kwargs = init_kwargs | {"scale": 1.0 * slam}
    return y, physical_models.BlackBody(**init_kwargs), norm, rescale


@u.quantity_input(y=slam, x=u.angstrom)
def fit_bb(
    bb_init: physical_models.BlackBody,
    y: slam,
    x: u.Quantity,
    rescaling: float,
    more_precise: bool = False,
    fitter: fitting._FitterMeta = fitting.LevMarLSQFitter,
) -> physical_models.BlackBody:

    fitter = fitter(calc_uncertainties=True)

    if more_precise:
        bb_fit = fitter(
            bb_init, x, y * rescaling, epsilon=1e-18, acc=1e-15, maxiter=1000
        )
    else:
        bb_fit = fitter(bb_init, x, y * rescaling)

    bb_fit.scale /= rescaling
    # bb_fit.scale.std /= rescaling

    return bb_fit


def get_bolo_flux(bb_fit: physical_models.BlackBody) -> u.Quantity:
    return bb_fit.bolometric_flux * np.pi


def calculate_radius(
    bb_scale: MyArrayLike, distance: u.Quantity = u.Quantity(10.0, unit=u.parsec)
) -> u.Quantity:
    """
    `distance` should be the luminosity distance if the blackbody was fitted to apparent flux, else it
    is 10 parsec, i.e. the absolute magnitude distance.
    """
    radius = bb_scale**0.5 * distance  # type: ignore
    return radius.to(u.cm)


def calculate_luminosity(bb_flux: u.Quantity, distance: u.Quantity) -> u.Quantity:
    lum: u.Quantity = (bb_flux * 4 * np.pi * distance**2).cgs  # type: ignore
    return lum.cgs


def get_parameters(
    bb_fit: physical_models.BlackBody,
    distance: u.Quantity = u.Quantity(10.0, unit=u.parsec),
) -> tuple[u.Quantity, u.Quantity, u.Quantity]:
    return (
        u.Quantity(bb_fit.temperature.quantity),
        calculate_radius(u.Quantity(bb_fit.scale.quantity), distance),
        get_bolo_flux(bb_fit),
    )


def fast_get_parameters(
    bb_fit: physical_models.BlackBody,
) -> tuple[float, float, float]:
    return bb_fit.temperature.value, bb_fit.scale.value, get_bolo_flux(bb_fit).value  # type: ignore


def bootstrap_mean_and_error(
    x: ArrayLike, axis: Optional[int] = None
) -> tuple[MyArrayLike, MyArrayLike]:
    return np.mean(x, axis=axis), np.std(x, axis=axis)


def safe_vary_flux(flux_varied: NDArray, flux_orig: NDArray) -> NDArray:
    return np.where(flux_varied > np.min(flux_varied) / 15, flux_varied, flux_orig)


def bootstrap_parameters(
    flux: u.Quantity,
    unc: u.Quantity,
    waves: u.Quantity,
    n_samples: int = 500,
    sigma_unc: float = 2,
    distance: u.Quantity = u.Quantity(10.0, unit=u.parsec),
    normalize_flux: bool = False,
) -> tuple[float]:
    """Return temperature, radius, and bolometric flux of a blackbody fit to a set of flux measurements.
    as well as their errors.
    """
    # npr = (np.random.random_sample(n_samples) - 0.5) * 2 * sigma_unc
    npr = (np.random.random_sample((n_samples, len(flux))) - 0.5) * 2 * sigma_unc

    flux_sr, bb_init, flux_normalization, rescaling = init_bb_fit(
        flux, normalize_flux=normalize_flux
    )

    unc_sr = unc / np.pi / u.sr
    _pars = []
    for n in npr:
        _flux_bs = flux_sr + n * unc_sr
        _flux_bs = safe_vary_flux(_flux_bs, flux_sr)
        bb_fit = fit_bb(bb_init, _flux_bs, waves, rescaling)
        if normalize_flux:
            bb_fit.scale *= flux_normalization
        _pars.append(fast_get_parameters(bb_fit))

    pars = np.array(_pars)
    pars[:, 1] = calculate_radius(pars[:, 1], distance=distance).value
    par_mean, par_se = bootstrap_mean_and_error(pars, axis=0)
    return tuple(np.hstack((par_mean, par_se)))


def multi_threaded_fit_with_bootstrap(
    fluxes: list[NDArray],
    flux_uncs: list[NDArray],
    waves: u.Quantity,
    n_samples: int = 500,
    sigma_unc: float = 2,
    distance: u.Quantity = u.Quantity(10.0, unit=u.parsec),
    normalize_flux: bool = False,
    threads: int = multiprocessing.cpu_count(),
) -> NDArray:

    if len(fluxes) == len(waves):
        flux_iter = [
            (f << flam, u << flam)
            for f, u in zip(*as_zipped_iterable_array(fluxes, flux_uncs))
        ]
        n_rows = fluxes[0].shape[0]
    else:
        flux_iter = [(f << flam, u << flam) for f, u in zip(fluxes, flux_uncs)]
        n_rows = len(fluxes)

        if len(waves) != len(fluxes[0]):
            raise ValueError(
                "Number of wavelengths and fluxes must be equal, and in same order."
            )

    pars = np.empty((n_rows, 8))

    partial_func = partial(
        bootstrap_parameters,
        waves=waves,
        distance=distance,
        normalize_flux=normalize_flux,
        sigma_unc=sigma_unc,
        n_samples=n_samples,
    )

    with multiprocessing.Pool(threads) as pool:
        with tqdm.tqdm(total=n_rows) as pbar:
            for i, par in enumerate(pool.starmap(partial_func, flux_iter)):
                pars[i] = (*par, 0, 0)
                pbar.update()

    pars[:, 6] = calculate_luminosity(pars[:, 2] * flux_units, distance=distance).value
    pars[:, 7] = calculate_luminosity(pars[:, 5] * flux_units, distance=distance).value
    return pars


def make_qtable(pars: NDArray) -> QTable:
    return QTable(
        data=pars,
        names=["T", "R", "F", "T_err", "R_err", "F_err", "L", "L_err"],
        units=(u.K, u.cm, flux_units, u.K, u.cm, flux_units, u.erg / u.s, u.erg / u.s),
    )


def as_zipped_iterable_array(
    fluxes: list[NDArray], flux_uncs: list[NDArray]
) -> tuple[NDArray, NDArray]:
    return make_zipped_stacked(fluxes), make_zipped_stacked(flux_uncs)


def make_zipped_stacked(list_of_arrays: list[NDArray]) -> NDArray:
    return np.vstack([array for array in list_of_arrays]).T


def load_pandas_flux_table(filename: str) -> tuple[list[NDArray], list[NDArray]]:
    fluxes = pd.read_json(filename)
    fluxes.index = fluxes.index.to_julian_date()  # type: ignore
    fluxes_zipped = list(
        fluxes[[name for name in fluxes.columns if not name.endswith("err")]].values
    )
    errors_zipped = list(
        fluxes[[name for name in fluxes.columns if name.endswith("err")]].values
    )
    return fluxes_zipped, errors_zipped


def load_phot(phot: Photometry, bands: dict[str, Filter],
              sorter: FilterSorter = FilterSorter()) -> tuple[u.Quantity, list[NDArray], list[NDArray], NDArray[float]]:
    bands = sorter.wave_sort(bands.values())

    fluxes = [b.name+'_flux' for b in bands]
    flux_uncs = [b.name+'_flux_err' for b in bands]
    waves = np.array([b.wave_eff.value for b in bands]) << u.AA

    df = phot.as_dataframe()
    df = df.pivot(index='jd', columns='band', values=['flux', 'flux_err'])
    df.columns = [f"{col[1]}_{col[0]}" for col in df.columns]
    df = df[fluxes + flux_uncs]

    fluxes_zipped = list(
        df[[name for name in df.columns if not name.endswith("err")]].values
    )
    errors_zipped = list(
        df[[name for name in df.columns if name.endswith("err")]].values
    )
    jds = df.index.values
    return waves, fluxes_zipped, errors_zipped, jds


def fit_sn(sn: SN, quasi: bool = False, blackbody: bool = False, n_samples: int = 1000,
           n_sigma: float = 2) -> SN:
    """
    Fit the SN with a blackbody and/or quasi-blackbody model.
    Returns a copy of the SN with the fitted values
    added to the photometry table, with sites and bands as "Bolo" and "BB".

    n_samples: number of bootstrap samples to use
    n_sigma: number of sigma to use for the bootstrap (e.g. vary errors by +/- 2 sigma)
    """
    waves, fluxes, flux_uncs, jds = load_phot(sn.phot, sn.bands)
    distance = sn.sninfo.cosmology.luminosity_distance(sn.sninfo.redshift).to(u.pc)
    new_sn = sn.copy()
    if not quasi and not blackbody:
        raise ValueError("Must fit at least one model. Set quasi=True or blackbody=True.")

    fac = PhotFactory(sn_phot=sn.phot)
    if quasi:
        # Setup
        fac.extend_phot(LumPhot)
        site_id = new_sn.sites.generate_id()
        new_sn.add_site(bolo_filt.name, id=site_id, marker='o')

        # Fit
        qbt = bootstrap_quasi_bolo(
            waves, np.array(fluxes), np.array(flux_uncs), n_samples=n_samples, sigma_unc=n_sigma,
            distance=distance)
        # Save

        qbt = qbt.to_pandas().rename({'F': 'flux', 'F_err': 'flux_err', 'L': 'lum', 'L_err': 'lum_err'}, axis=1)
        qbt['jd'] = jds
        qbt['band'] = bolo_filt.name
        qbt['site'] = site_id
        fac.concat_phot(fac.from_df(qbt))

    if blackbody:
        # Setup
        fac.extend_phot(BBLumPhot)
        site_id = new_sn.sites.generate_id()
        new_sn.add_site(blackbody_filt.name, id=site_id, marker='s')

        # Fit
        print("Fitting blackbody to flux measurements")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fitted_params = multi_threaded_fit_with_bootstrap(
                waves=waves,
                fluxes=fluxes,
                flux_uncs=flux_uncs,
                n_samples=n_samples,
                sigma_unc=n_sigma,
                distance=distance,  # type: ignore
                normalize_flux=False,
                threads=max(multiprocessing.cpu_count()-3, 1),
            )

        # Save
        print("Done: Saving results!")
        qt = make_qtable(fitted_params).to_pandas()
        qt = qt.rename({'F': 'flux', 'F_err': 'flux_err', 'L': 'lum', 'L_err': 'lum_err',
                        'R': 'radius', 'R_err': 'radius_err', 'T': 'temp', 'T_err': 'temp_err'},
                       axis=1)
        qt['jd'] = jds
        qt['band'] = blackbody_filt.name
        qt['site'] = site_id
        fac.concat_phot(fac.from_df(qt))

    new_sn.phot = fac.sn_phot
    new_sn.set_phases()
    new_sn.make_bands(new_sn.phot.band.unique().tolist(), ebv=sn.sninfo.ebv)

    return new_sn


def bootstrap_quasi_bolo(
    waves: NDArray,
    fluxes: NDArray,
    flux_errs: NDArray,
    n_samples: int = 500,
    sigma_unc: float = 2,
    distance: u.Quantity = u.Quantity(10.0, unit=u.parsec),
) -> QTable:
    fluxes = fluxes << flam
    flux_errs = flux_errs << flam
    min_flux = fluxes.min() / 10
    bs_samples = (np.random.random((n_samples, *fluxes.shape)) - 0.5) * 2 * sigma_unc * flux_errs + fluxes
        
    bs_samples[bs_samples < 0] = min_flux
    bs_samples = np.trapz(bs_samples, waves) 
    bs_lums = calculate_luminosity(bs_samples, distance=distance)

    return QTable(
        data={
            "F": bs_samples.mean(axis=0),
            "F_err": bs_samples.std(axis=0),
            "L": bs_lums.mean(axis=0),
            "L_err": bs_lums.std(axis=0),
        }
    )


def main():
    pass
    # waves = np.array([4299.2, 4900.139, 6241.273, 7563.767, 5393.9]) * u.AA
    # fluxes, flux_uncs = load_pandas_flux_table(
    #     "/home/emir/Dropbox/SN2020lao/emir/20lao/Tables/absflux_per_filter.json"
    # )
    # redshift = 0.03112
    # cosmology = WMAP5
    # distance = cosmology.luminosity_distance(redshift)
    # absolute = True
    # if absolute:
    #     distance = u.Quantity(10.0, unit=u.parsec)
    #
    # qbt = bootstrap_quasi_bolo(
    #     waves, np.array(fluxes), np.array(flux_uncs), n_samples=1000, sigma_unc=2,
    # distance=distance)
    # qbt.write(
    #     "/home/emir/Dropbox/SN2020lao/emir/20lao/Tables/quasi_bolometric_flux_lum.ecsv",
    #     format="ascii.ecsv",
    #     overwrite=True,
    # )
    #
    # print("Fitting blackbody to flux measurements")
    #
    # with warnings.catch_warnings():
    #     warnings.simplefilter("ignore")
    #     fitted_params = multi_threaded_fit_with_bootstrap(
    #         waves=waves,
    #         fluxes=fluxes,
    #         flux_uncs=flux_uncs,
    #         n_samples=1000,
    #         sigma_unc=3,
    #         distance=distance,  # type: ignore
    #         normalize_flux = False,
    #         threads= 12
    #     )
    # print("Done: Saving results!")
    #
    # qt = make_qtable(fitted_params)
    # qt.write("/home/emir/Dropbox/SN2020lao/emir/20lao/Tables/bolo_from_bb_1000_3.ecsv", overwrite=True, format="ascii.ecsv")
    #
    # with warnings.catch_warnings():
    #     warnings.simplefilter("ignore")
    #     fitted_params = multi_threaded_fit_with_bootstrap(
    #         waves=waves,
    #         fluxes=fluxes,
    #         flux_uncs=flux_uncs,
    #         n_samples=500,
    #         sigma_unc=2,
    #         distance=distance,  # type: ignore
    #         normalize_flux = False,
    #         threads= 12
    #     )
    # print("Done: Saving results!")
    #
    # qt = make_qtable(fitted_params)
    # qt.write("/home/emir/Dropbox/SN2020lao/emir/20lao/Tables/bolo_from_bb_500_2.ecsv", overwrite=True, format="ascii.ecsv")
    #
    # with warnings.catch_warnings():
    #     warnings.simplefilter("ignore")
    #     fitted_params = multi_threaded_fit_with_bootstrap(
    #         waves=waves,
    #         fluxes=fluxes,
    #         flux_uncs=flux_uncs,
    #         n_samples=1000,
    #         sigma_unc=2,
    #         distance=distance,  # type: ignore
    #         normalize_flux = False,
    #         threads= 12
    #     )
    # print("Done: Saving results!")
    #
    # qt = make_qtable(fitted_params)
    # qt.write("/home/emir/Dropbox/SN2020lao/emir/20lao/Tables/bolo_from_bb_1000_2.ecsv", overwrite=True, format="ascii.ecsv")


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import Optional, Protocol, ClassVar
import numpy as np
import pandas as pd


# noinspection PyUnusedLocal
@dataclass(frozen=True)
class Converter(Protocol):
    glob_str: ClassVar[str]
    jd_col: str
    mag_col: str
    mag_err_col: str
    band_col: str
    sub_col: str
    site_col: str
    df: pd.DataFrame

    def __init__(self, df: pd.DataFrame) -> None:
        ...

    def jd_convert(self) -> None:
        ...

    def mag_convert(self) -> None:
        ...

    def mag_err_convert(self) -> None:
        ...

    def band_convert(self) -> None:
        ...

    def sub_convert(self) -> None:
        ...

    def site_convert(self) -> None:
        ...

    def convert(self) -> pd.DataFrame:
        ...


@dataclass(frozen=True)
class BaseConverter(Converter):
    glob_str: ClassVar[str] = '*'
    jd_col: str = 'time'
    mag_col: str = 'mag'
    mag_err_col: str = 'mag_err'
    band_col: str = 'filter'
    sub_col: str = 'sub'
    site_col: str = 'site'
    df: pd.DataFrame = field(default_factory=pd.DataFrame)

    def jd_convert(self) -> None:
        self.df['jd'] = self.df[self.jd_col]

    def mag_convert(self) -> None:
        self.df['mag'] = self.df[self.mag_col]

    def mag_err_convert(self) -> None:
        self.df['mag_err'] = self.df[self.mag_err_col]

    def band_convert(self) -> None:
        self.df['band'] = self.df[self.band_col]

    def sub_convert(self) -> None:
        self.df['sub'] = self.df[self.sub_col]

    def site_convert(self) -> None:
        self.df['site'] = self.df[self.site_col]

    def convert(self) -> pd.DataFrame:
        self.jd_convert()
        self.mag_convert()
        self.mag_err_convert()
        self.band_convert()
        self.sub_convert()
        self.site_convert()
        return self.df


@dataclass(frozen=True)
class GenericCSVConverter(BaseConverter):
    glob_str: ClassVar[str] = '*.csv'

    def sub_convert(self) -> None:
        self.df['sub'] = True

    def site_convert(self) -> None:
        self.df['site'] = np.random.randint(10, 100)


@dataclass(frozen=True)
class GenericECSVConverter(BaseConverter):
    glob_str: ClassVar[str] = '*.ecsv'

    def sub_convert(self) -> None:
        self.df['sub'] = True

    def site_convert(self) -> None:
        self.df['site'] = np.random.randint(10, 100)


@dataclass(frozen=True)
class GenericJSONConverter(BaseConverter):
    glob_str: ClassVar[str] = '*.json'

    def sub_convert(self) -> None:
        self.df['sub'] = True

    def site_convert(self) -> None:
        self.df['site'] = np.random.randint(10, 100)



from .ingest import FlowsIngest, DataIngestor, GenericIngest, CSVIngest, ECSVIngest, JSONIngest, Ingestors, ingest_sn
from .collators import collate_csv, collate_json, collate_ecsv
from .readers import read_astropy_table, read_pandas_csv, read_astropy_table_as_df_with_times_as_jd



from astrosn.ingest import ingest_sn, Ingestors
from astrosn import SN, Site
import astrosn.plotting as plot
from astrosn.plotting.plot_lc import _get_legend_handles_labels
import matplotlib.pyplot as plt

def test_flows_ingest() -> SN:
    sn = ingest_sn("inputs/2021aess/", "SN2021aess", "flows")
    assert sn.name == "SN2021aess"
    assert sn.redshift > 0.0
    ztf_site = sn.sites["ZTF"]
    assert isinstance(ztf_site, Site)
    assert ztf_site.name == "ZTF"
    assert ztf_site.id == 0
    assert len(sn.phot) > 0

    sn = ingest_sn("inputs/2021aess/", "SN2021aess", Ingestors.flows)
    assert sn.name == "SN2021aess"
    assert sn.redshift > 0.0
    ztf_site = sn.sites["ZTF"]
    assert isinstance(ztf_site, Site)
    assert ztf_site.name == "ZTF"
    assert ztf_site.id == 0
    assert len(sn.phot) > 0

    return sn


if __name__ == "__main__":
    sn = test_flows_ingest()
    sn.sub_only = False
    sn.sninfo.sub_only = False
    sn = sn.restframe()

    plotpars = plot.AbsMagPlotParams
    plotpars.label_sites = False
    plotpars.split_by_site = False
    fig, ax = plot.plot_flows_absmag(sn, shift=False)
    handles, labels = _get_legend_handles_labels(ax.legend_)
    plt.show(block=True)
    print(labels, len(labels))

    plotpars = plot.AbsMagPlotParams
    plotpars.label_sites = True
    plotpars.split_by_site = True
    fig, ax = plot.plot_flows_absmag(sn, shift=False)
    handles, labels = _get_legend_handles_labels(ax.legend_)
    print(labels, len(labels))

    print(len(handles), len(labels))
